#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from copy import deepcopy
from typing import Dict, Iterable, List, NamedTuple, Optional, Set, Tuple

import numpy as np
from ax.core.arm import Arm
from ax.core.base_trial import BaseTrial, TrialStatus
from ax.core.batch_trial import BatchTrial
from ax.core.data import Data
from ax.core.experiment import Experiment
from ax.core.generator_run import GeneratorRun
from ax.core.objective import MultiObjective
from ax.core.observation import ObservationFeatures
from ax.core.optimization_config import OptimizationConfig
from ax.core.trial import Trial
from ax.core.types import ComparisonOp
from pyre_extensions import none_throws

TArmTrial = Tuple[str, int]

# Threshold for switching to pending points extraction based on trial status.
MANY_TRIALS_IN_EXPERIMENT = 100


# --------------------------- Data intergrity utils. ---------------------------


class MissingMetrics(NamedTuple):
    objective: Dict[str, Set[TArmTrial]]
    outcome_constraints: Dict[str, Set[TArmTrial]]
    tracking_metrics: Dict[str, Set[TArmTrial]]


def get_missing_metrics(
    data: Data, optimization_config: OptimizationConfig
) -> MissingMetrics:
    """Return all arm_name, trial_index pairs, for which some of the
    observatins of optimization config metrics are missing.

    Args:
        data: Data to search.
        optimization_config: provides metric_names to search for.

    Returns:
        A NamedTuple(missing_objective, Dict[str, missing_outcome_constraint])
    """
    objective = optimization_config.objective
    if isinstance(objective, MultiObjective):
        objective_metric_names = [m.name for m in objective.metrics]
    else:
        objective_metric_names = [optimization_config.objective.metric.name]

    outcome_constraints_metric_names = [
        outcome_constraint.metric.name
        for outcome_constraint in optimization_config.outcome_constraints
    ]
    missing_objectives = {
        objective_metric_name: _get_missing_arm_trial_pairs(data, objective_metric_name)
        for objective_metric_name in objective_metric_names
    }
    missing_outcome_constraints = get_missing_metrics_by_name(
        data, outcome_constraints_metric_names
    )
    all_metric_names = set(data.df["metric_name"])
    optimization_config_metric_names = set(missing_objectives.keys()).union(
        outcome_constraints_metric_names
    )
    missing_tracking_metric_names = all_metric_names.difference(
        optimization_config_metric_names
    )
    missing_tracking_metrics = get_missing_metrics_by_name(
        data=data, metric_names=missing_tracking_metric_names
    )
    return MissingMetrics(
        objective={k: v for k, v in missing_objectives.items() if len(v) > 0},
        outcome_constraints={
            k: v for k, v in missing_outcome_constraints.items() if len(v) > 0
        },
        tracking_metrics={
            k: v for k, v in missing_tracking_metrics.items() if len(v) > 0
        },
    )


def get_missing_metrics_by_name(
    data: Data, metric_names: Iterable[str]
) -> Dict[str, Set[TArmTrial]]:
    """Return all arm_name, trial_index pairs missing some observations of
    specified metrics.

    Args:
        data: Data to search.
        metric_names: list of metrics to search for.

    Returns:
        A Dict[str, missing_metrics], one entry for each metric_name.
    """
    missing_metrics = {
        metric_name: _get_missing_arm_trial_pairs(data=data, metric_name=metric_name)
        for metric_name in metric_names
    }
    return missing_metrics


def _get_missing_arm_trial_pairs(data: Data, metric_name: str) -> Set[TArmTrial]:
    """Return arm_name and trial_index pairs missing a specified metric."""
    metric_df = data.df[data.df.metric_name == metric_name]
    present_metric_df = metric_df[metric_df["mean"].notnull()]
    arm_trial_pairs = set(zip(data.df["arm_name"], data.df["trial_index"]))
    arm_trial_pairs_with_metric = set(
        zip(present_metric_df["arm_name"], present_metric_df["trial_index"])
    )
    missing_arm_trial_pairs = arm_trial_pairs.difference(arm_trial_pairs_with_metric)
    return missing_arm_trial_pairs


# -------------------- Experiment result extraction utils. ---------------------


def best_feasible_objective(
    optimization_config: OptimizationConfig, values: Dict[str, np.ndarray]
) -> np.ndarray:
    """Compute the best feasible objective value found by each iteration.

    Args:
        optimization_config: Optimization config.
        values: Dictionary from metric name to array of value at each
            iteration. If optimization config contains outcome constraints, values
            for them must be present in `values`.

    Returns: Array of cumulative best feasible value.
    """
    # Get objective at each iteration
    objective = optimization_config.objective
    f = values[objective.metric.name]
    # Set infeasible points to have infinitely bad values
    infeas_val = np.inf if objective.minimize else -np.inf
    for oc in optimization_config.outcome_constraints:
        if oc.relative:
            raise ValueError(
                "Benchmark aggregation does not support relative constraints"
            )
        g = values[oc.metric.name]
        feas = g <= oc.bound if oc.op == ComparisonOp.LEQ else g >= oc.bound
        f[~feas] = infeas_val

    # Get cumulative best
    minimize = objective.minimize
    accumulate = np.minimum.accumulate if minimize else np.maximum.accumulate
    return accumulate(f)


def _extract_generator_runs(trial: BaseTrial) -> List[GeneratorRun]:
    if isinstance(trial, BatchTrial):
        return trial.generator_runs
    if isinstance(trial, Trial):
        return [none_throws(trial.generator_run)]
    raise ValueError("Unexpected trial type")


def get_model_trace_of_times(
    experiment: Experiment,
) -> Tuple[List[Optional[float]], List[Optional[float]]]:
    """
    Get time spent fitting the model and generating candidates during each trial.
    Not cumulative.

    Returns:
        List of fit times, list of gen times.
    """
    generator_runs = [
        gr
        for trial in experiment.trials.values()
        for gr in _extract_generator_runs(trial=trial)
    ]
    fit_times = [gr.fit_time for gr in generator_runs]
    gen_times = [gr.gen_time for gr in generator_runs]
    return fit_times, gen_times


def get_model_times(experiment: Experiment) -> Tuple[float, float]:
    """Get total times spent fitting the model and generating candidates in the
    course of the experiment.
    """
    fit_times, gen_times = get_model_trace_of_times(experiment)
    fit_time = sum((t for t in fit_times if t is not None))
    gen_time = sum((t for t in gen_times if t is not None))
    return fit_time, gen_time


# -------------------- Pending observations extraction utils. ---------------------


def extract_pending_observations(
    experiment: Experiment,
    include_out_of_design_points: bool = False,
) -> Optional[Dict[str, List[ObservationFeatures]]]:
    """Computes a list of pending observation features (corresponding to:
    - arms that have been generated and run in the course of the experiment,
    but have not been completed with data,
    - arms that have been abandoned or belong to abandoned trials).

    This function dispatches to:
    - ``get_pending_observation_features`` if experiment is using
    ``BatchTrial``-s or has fewer than 100 trials,
    - ``get_pending_observation_features_based_on_trial_status`` if
    experiment is using  ``Trial``-s and has more than 100 trials.

    ``get_pending_observation_features_based_on_trial_status`` is a faster
    way to compute pending observations, but it is not guaranteed to be
    accurate for ``BatchTrial`` settings and makes assumptions, e.g.
    arms in ``COMPLETED`` trial never being pending. See docstring of
    that function for more details.

    NOTE: Pending observation features are passed to the model to
    instruct it to not generate the same points again.
    """
    if len(experiment.trials) >= MANY_TRIALS_IN_EXPERIMENT and all(
        isinstance(t, Trial) for t in experiment.trials.values()
    ):
        return get_pending_observation_features_based_on_trial_status(
            experiment=experiment,
            include_out_of_design_points=include_out_of_design_points,
        )

    return get_pending_observation_features(
        experiment=experiment, include_out_of_design_points=include_out_of_design_points
    )


def get_pending_observation_features(
    experiment: Experiment,
    *,
    include_out_of_design_points: bool = False,
) -> Optional[Dict[str, List[ObservationFeatures]]]:
    """Computes a list of pending observation features (corresponding to:
    - arms that have been generated in the course of the experiment,
    but have not been completed with data,
    - arms that have been abandoned or belong to abandoned trials).

    NOTE: Pending observation features are passed to the model to
    instruct it to not generate the same points again.

    Args:
        experiment: Experiment, pending features on which we seek to compute.
        include_out_of_design_points: By default, this function will not include
            "out of design" points (those that are not in the search space) among
            the pending points. This is because pending points are generally used to
            help the model avoid re-suggesting the same points again. For points
            outside of the search space, this will not happen, so they typically do
            not need to be included. However, if the user wants to include them,
            they can be included by setting this flag to ``True``.

    Returns:
        An optional mapping from metric names to a list of observation features,
        pending for that metric (i.e. do not have evaluation data for that metric).
        If there are no pending features for any of the metrics, return is None.
    """

    def _is_in_design(arm: Arm) -> bool:
        return experiment.search_space.check_membership(parameterization=arm.parameters)

    pending_features = {metric_name: [] for metric_name in experiment.metrics}

    def create_observation_feature(
        arm: Arm,
        trial_index: int,
        trial: BaseTrial,
    ) -> Optional[ObservationFeatures]:
        if not include_out_of_design_points and not _is_in_design(arm=arm):
            return None
        return ObservationFeatures.from_arm(
            arm=arm,
            trial_index=trial_index,
            metadata=trial._get_candidate_metadata(arm_name=arm.name),
        )

    # Note that this assumes that if a metric appears in fetched data, the trial is
    # not pending for the metric. Where only the most recent data matters, this will
    # work, but may need to add logic to check previously added data objects, too.
    for trial_index, trial in experiment.trials.items():
        if trial.status.is_deployed:
            metric_names_in_data = set(trial.lookup_data().df.metric_name.values)
        else:
            metric_names_in_data = set()

        for metric_name in experiment.metrics:
            if metric_name not in pending_features:
                pending_features[metric_name] = []

            if (
                trial.status.is_abandoned
                or trial.status.is_candidate
                or (
                    trial.status.is_deployed
                    and metric_name not in metric_names_in_data
                    and trial.arms is not None
                )
            ):
                for arm in trial.arms:
                    if feature := create_observation_feature(
                        arm=arm,
                        trial_index=trial_index,
                        trial=trial,
                    ):
                        pending_features[metric_name].append(feature)

            # Also add abandoned arms as pending for all metrics.
            if isinstance(trial, BatchTrial):
                for arm in trial.abandoned_arms:
                    if feature := create_observation_feature(
                        arm=arm,
                        trial_index=trial_index,
                        trial=trial,
                    ):
                        pending_features[metric_name].append(feature)

    return pending_features if any(x for x in pending_features.values()) else None


# TODO: allow user to pass search space which overrides that on the experiment
# (to use for the `include_out_of_design_points` check)
def get_pending_observation_features_based_on_trial_status(
    experiment: Experiment,
    include_out_of_design_points: bool = False,
) -> Optional[Dict[str, List[ObservationFeatures]]]:
    """A faster analogue of ``get_pending_observation_features`` that makes
    assumptions about trials in experiment in order to speed up extraction
    of pending points.

    Assumptions:

    * All arms in all trials in ``CANDIDATE``, ``STAGED``, ``RUNNING`` and ``ABANDONED``
      statuses are to be considered pending for all outcomes.
    * All arms in all trials in other statuses are to be considered not pending for
      all outcomes.

    This entails:

    * No actual data-fetching for trials to determine whether arms in them are pending
      for specific outcomes.
    * Even if data is present for some outcomes in ``RUNNING`` trials, their arms will
      still be considered pending for those outcomes.

    NOTE: This function should not be used to extract pending features in field
    experiments, where arms in running trials should not be considered pending if
    there is data for those arms.

    Args:
        experiment: Experiment, pending features on which we seek to compute.

    Returns:
        An optional mapping from metric names to a list of observation features,
        pending for that metric (i.e. do not have evaluation data for that metric).
        If there are no pending features for any of the metrics, return is None.
    """
    pending_features_list = []
    for status in [
        TrialStatus.CANDIDATE,
        TrialStatus.STAGED,
        TrialStatus.RUNNING,
        TrialStatus.ABANDONED,
    ]:
        for trial in experiment.trials_by_status[status]:
            for arm in trial.arms:
                if (
                    include_out_of_design_points
                    or experiment.search_space.check_membership(arm.parameters)
                ):
                    pending_features_list.append(
                        ObservationFeatures.from_arm(
                            arm=arm,
                            trial_index=trial.index,
                            metadata=trial._get_candidate_metadata(arm_name=arm.name),
                        )
                    )
    pending_features = {
        # Using deepcopy here to avoid issues due to in-place transforms.
        metric_name: deepcopy(pending_features_list)
        for metric_name in experiment.metrics
    }
    return pending_features if pending_features_list else None


def extend_pending_observations(
    experiment: Experiment,
    pending_observations: Dict[str, List[ObservationFeatures]],
    generator_run: GeneratorRun,
) -> None:
    """Extend given pending observations dict (from metric name to observations
    that are pending for that metric), with arms in a given generator run.
    """
    for m in experiment.metrics:
        if m not in pending_observations:
            pending_observations[m] = []
        pending_observations[m].extend(
            ObservationFeatures.from_arm(a) for a in generator_run.arms
        )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from collections import defaultdict
from typing import Dict, List, Tuple

from ax.core.search_space import SearchSpace
from ax.core.types import TParameterization

from ax.utils.common.typeutils import checked_cast


def search_space_update_recommendation(
    search_space: SearchSpace,
    parametrizations: List[TParameterization],
    boundary_prop: float = 0.5,
    tol: float = 1e-6,
) -> Tuple[Dict[str, Tuple[float, float]], str]:
    r"""
    Recommendation to update the search space boundaries depending on the
    generated candidates. If most of them fall on some of the boundaries, we
    suggest expanding the search space along those boundaries.

    Note that this function does not handle parameter constraints.
    TODO: Add support for search space update in the presence of
    paramater constraints.

    Args:
        search_space: Search space.
        parametrizations: A list of suggested parametrizations (parameter values).
        boundary_prop: The minimal proportion of suggested parametrizations
            that land at the boundary of the search space for us to recommend
            expanding the search space.
        tol: Tolerance for the difference between parameters and the boundary
            lower and upper bound.

    Returns: A tuple consisting of
        - a dictionary mapping parameter names into the proportion of suggested
            parameterizations that landed on the lower and upper bound of
            that parameter respectively and
        - a human readable string containing the recommendation on the
            search space expansion to expand the boundaries
            where more than a specified proportion of parametrizations land.
    """
    # mapping from range parameter names to the proportions of suggested
    # parametrizations that are at that parameter lower and upper boundary,
    # e.g., {"a": (0.1, 0.2)} means 10% of the suggested parametrizations
    # have parameter "a" value equal to a.lower and 20% have parameter "a" value
    # equal to a.upper.
    param_boundary_prop = defaultdict()
    msg = str()

    num_suggestions = len(parametrizations)

    for parameter_name, parameter in search_space.range_parameters.items():
        lb = 0  # counts how many parameters are equal to the boundary's lower bound
        ub = 0  # counts how many parameters are equal to the boundary's lower bound
        for parametrization in parametrizations:
            value = parametrization[parameter_name]
            value = checked_cast(float, value)
            if abs(value - parameter.lower) < tol:
                lb += 1
            elif abs(value - parameter.upper) < tol:
                ub += 1

        prob_lower = lb / float(num_suggestions)
        prob_upper = ub / float(num_suggestions)
        param_boundary_prop[parameter_name] = (prob_lower, prob_upper)

        if prob_lower >= boundary_prop:
            msg += (
                f"\n Parameter {parameter_name} values are at the lower bound "
                f"({parameter.lower}) of the search space "
                f"for at least {boundary_prop * 100}% of all suggested parameters. "
                "Consider decreasing this lower bound on the search space and "
                "re-generating the candidates inside the expanded search space. "
            )

        if prob_upper >= boundary_prop:
            msg += (
                f"\n Parameter {parameter_name} values are at the upper bound "
                f"({parameter.upper}) of the search space "
                f"for at last {boundary_prop * 100}% of all suggested parameters. "
                "Consider increasing this upper bound on the search space and "
                "re-generating the candidates inside the expanded search space. "
            )

    return param_boundary_prop, msg



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from typing import Any, Dict, Set

import numpy as np

import pandas as pd

from ax.analysis.base_plotly_visualization import BasePlotlyVisualization

from ax.analysis.helpers.layout_helpers import updatemenus_format

from ax.analysis.helpers.plot_data_df_helpers import get_plot_data_in_sample_arms_df
from ax.analysis.helpers.plot_helpers import arm_name_to_sort_key, resize_subtitles

from ax.analysis.helpers.scatter_helpers import (
    error_dot_plot_trace_from_df,
    relativize_dataframe,
)

from ax.core.experiment import Experiment

from ax.exceptions.core import UnsupportedPlotError

from ax.modelbridge import ModelBridge

from plotly import graph_objs as go


class PredictedOutcomesDotPlot(BasePlotlyVisualization):
    def __init__(
        self,
        experiment: Experiment,
        model: ModelBridge,
    ) -> None:
        """
        Args:
            experiment: The experiment associated with this plot
            model: model which is used to fetch the plotting data.
        """

        self.model = model
        self.metrics: Set[str] = model.metric_names
        if model.status_quo is None or model.status_quo.arm_name is None:
            raise UnsupportedPlotError(
                "status quo must be specified for PredictedOutcomesDotPlot"
            )
        self.status_quo_name: str = model.status_quo.arm_name

        super().__init__(experiment=experiment)

    def get_df(self) -> pd.DataFrame:
        """
        Returns:
            A dataframe containing:
            {
                "arm_name": name of the arm in the cross validation result
                "metric_name": name of the observed/predicted metric
                "x": value of the observation for the metric for this arm
                "x_se": standard error of the observation for the metric of this arm
                "y": value predicted for the metric for this arm
                "y_se": standard error of predicted metric for this arm
                "arm_parameters": Parametrization of the arm
                "rel": whether the data is relativized with respect to status quo
            }"""
        return relativize_dataframe(
            get_plot_data_in_sample_arms_df(
                model=self.model, metric_names=self.metrics
            ),
            status_quo_name=self.status_quo_name,
        )

    def get_fig(
        self,
    ) -> go.Figure:
        """
        For each metric, we plot the predicted values for each arm along with its CI
        These values are relativized with respect to the status quo.
        """
        name_order_axes: Dict[str, Dict[str, Any]] = {}

        in_sample_df = self.get_df()
        traces = []
        metric_dropdown = []

        for i, metric in enumerate(self.metrics):
            filtered_df = in_sample_df.loc[in_sample_df["metric_name"] == metric]
            data_single: Dict[str, Any] = error_dot_plot_trace_from_df(
                df=filtered_df, show_CI=True, visible=(i == 0)
            )

            # order arm name sorting arm numbers within batch
            names_by_arm = sorted(
                np.unique(data_single["x"]),
                key=lambda x: arm_name_to_sort_key(x),
                reverse=True,
            )

            name_order_axes["xaxis{}".format(i + 1)] = {
                "categoryorder": "array",
                "categoryarray": names_by_arm,
                "type": "category",
            }
            name_order_axes["yaxis{}".format(i + 1)] = {
                "ticksuffix": "%",
                "zerolinecolor": "red",
            }

            traces.append(data_single)

            is_visible = [False] * (len(metric))
            is_visible[i] = True
            metric_dropdown.append(
                {
                    "args": [
                        {
                            "visible": is_visible,
                        },
                    ],
                    "label": metric,
                    "method": "update",
                }
            )

        updatemenus = updatemenus_format(metric_dropdown=metric_dropdown)

        fig = go.Figure(data=traces)

        fig["layout"].update(
            updatemenus=updatemenus,
            width=1030,
            height=500,
            **name_order_axes,
        )

        fig = resize_subtitles(figure=fig, size=10)
        fig["layout"]["title"] = "Predicted Outcomes by Metric"
        return fig



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from unittest import mock

import numpy as np
from ax.core.data import Data
from ax.core.experiment import Experiment
from ax.core.observation import Observation, ObservationData, ObservationFeatures
from ax.core.parameter import ParameterType, RangeParameter
from ax.core.search_space import SearchSpace
from ax.modelbridge.base import ModelBridge
from ax.modelbridge.transforms.utils import (
    ClosestLookupDict,
    derelativize_optimization_config_with_raw_status_quo,
)
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import get_multi_objective_optimization_config

OBSERVATION_DATA = [
    Observation(
        features=ObservationFeatures(parameters={"x": 2.0, "y": 10.0}),
        data=ObservationData(
            means=np.array([1.0, 2.0, 6.0]),
            covariance=np.array([[1.0, 2.0, 0.0], [3.0, 4.0, 0.0], [0.0, 0.0, 4.0]]),
            metric_names=["m1", "m2", "m3"],
        ),
        arm_name="1_1",
    )
]


class TransformUtilsTest(TestCase):
    def test_closest_lookup_dict(self) -> None:
        # test empty lookup
        d = ClosestLookupDict()
        with self.assertRaises(RuntimeError):
            d[0]
        # basic test
        keys = (1.0, 2, 4)
        vals = ("a", "b", "c")
        d = ClosestLookupDict(zip(keys, vals))
        for k, v in zip(keys, vals):
            self.assertEqual(d[k], v)
        self.assertEqual(d[2.5], "b")
        self.assertEqual(d[0], "a")
        self.assertEqual(d[6], "c")
        with self.assertRaises(ValueError):
            # pyre-fixme[6]: For 1st param expected `Number` but got `str`.
            d["str_key"] = 3

    @mock.patch(
        "ax.modelbridge.base.observations_from_data",
        autospec=True,
        return_value=(OBSERVATION_DATA),
    )
    def test_derelativize_optimization_config_with_raw_status_quo(self, _) -> None:
        optimization_config = get_multi_objective_optimization_config()
        dummy_search_space = SearchSpace(
            parameters=[
                RangeParameter("x", ParameterType.FLOAT, 0, 20),
                RangeParameter("y", ParameterType.FLOAT, 0, 20),
            ]
        )
        modelbridge = ModelBridge(
            search_space=dummy_search_space,
            model=None,
            transforms=[],
            experiment=Experiment(dummy_search_space, "test"),
            data=Data(),
            optimization_config=optimization_config,
            status_quo_name="1_1",
        )
        new_opt_config = derelativize_optimization_config_with_raw_status_quo(
            optimization_config=optimization_config,
            modelbridge=modelbridge,
            observations=OBSERVATION_DATA,
        )
        expected_bound_values = {"m1": 0.9975, "m2": 1.995, "m3": 5.985}
        for oc in new_opt_config.all_constraints:
            self.assertFalse(oc.relative)
            expected_bound_value = expected_bound_values[oc.metric.name]
            self.assertEqual(oc.bound, expected_bound_value)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from typing import Any, Iterable


class DataProviderError(Exception):
    """Base Exception for Ax DataProviders.

    The type of the data provider must be included.
    The raw error is stored in the data_provider_error section,
    and an Ax-friendly message is stored as the actual error message.
    """

    def __init__(
        self,
        message: str,
        data_provider: str,
        # pyre-fixme[2]: Parameter annotation cannot be `Any`.
        data_provider_error: Any,
    ) -> None:
        self.message = message
        self.data_provider = data_provider
        self.data_provider_error = data_provider_error

    def __str__(self) -> str:
        return (
            "{message}. \n Error thrown by: {dp} data provider \n"
            + "Native {dp} data provider error: {dp_error}"
        ).format(
            dp=self.data_provider,
            message=self.message,
            dp_error=self.data_provider_error,
        )


class MissingDataError(Exception):
    def __init__(self, missing_trial_indexes: Iterable[int]) -> None:
        missing_trial_str = ", ".join([str(index) for index in missing_trial_indexes])
        self.message: str = (
            f"Unable to find data for the following trials: {missing_trial_str} "
            "consider updating the data fetching kwargs or manually fetching "
            "data via `refetch_data()`"
        )

    def __str__(self) -> str:
        return self.message



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import copy
import itertools
from collections import namedtuple
from logging import INFO, WARN
from typing import Dict, List
from unittest import mock
from unittest.mock import patch

import pandas as pd
from ax.core.arm import Arm
from ax.core.metric import Metric
from ax.core.objective import MultiObjective, Objective
from ax.core.optimization_config import (
    MultiObjectiveOptimizationConfig,
    OptimizationConfig,
)
from ax.core.outcome_constraint import ObjectiveThreshold
from ax.core.types import ComparisonOp
from ax.modelbridge.generation_node import GenerationStep
from ax.modelbridge.generation_strategy import GenerationStrategy
from ax.modelbridge.registry import Models
from ax.service.scheduler import Scheduler
from ax.service.utils.report_utils import (
    _format_comparison_string,
    _get_cross_validation_plots,
    _get_curve_plot_dropdown,
    _get_metric_name_pairs,
    _get_objective_trace_plot,
    _get_objective_v_param_plots,
    _get_shortest_unique_suffix_dict,
    _objective_vs_true_objective_scatter,
    BASELINE_ARM_NAME,
    compare_to_baseline,
    compute_maximum_map_values,
    exp_to_df,
    Experiment,
    FEASIBLE_COL_NAME,
    get_standard_plots,
    plot_feature_importance_by_feature_plotly,
    select_baseline_arm,
    warn_if_unpredictable_metrics,
)
from ax.service.utils.scheduler_options import SchedulerOptions
from ax.utils.common.testutils import TestCase
from ax.utils.common.typeutils import checked_cast, not_none
from ax.utils.testing.core_stubs import (
    get_branin_experiment,
    get_branin_experiment_with_multi_objective,
    get_branin_experiment_with_timestamp_map_metric,
    get_branin_search_space,
    get_experiment_with_observations,
    get_high_dimensional_branin_experiment,
    get_multi_type_experiment,
    get_test_map_data_experiment,
)
from ax.utils.testing.mock import fast_botorch_optimize
from ax.utils.testing.modeling_stubs import get_generation_strategy
from plotly import graph_objects as go

OBJECTIVE_NAME = "branin"
PARAMETER_COLUMNS = ["x1", "x2"]
FLOAT_COLUMNS: List[str] = [OBJECTIVE_NAME] + PARAMETER_COLUMNS
EXPECTED_COLUMNS: List[str] = [
    "trial_index",
    "arm_name",
    "trial_status",
    "generation_method",
] + FLOAT_COLUMNS
DUMMY_OBJECTIVE_MEAN = 1.2345
DUMMY_SOURCE = "test_source"
DUMMY_MAP_KEY = "test_map_key"
TRUE_OBJECTIVE_NAME = "other_metric"
TRUE_OBJECTIVE_MEAN = 2.3456
DUMMY_MSG = "test_message"


class ReportUtilsTest(TestCase):
    @patch(
        "ax.service.utils.report_utils._merge_results_if_no_duplicates",
        autospec=True,
        return_value=pd.DataFrame(
            [
                # Trial indexes are out-of-order.
                {"arm_name": "a", "trial_index": 1},
                {"arm_name": "b", "trial_index": 2},
                {"arm_name": "c", "trial_index": 0},
            ]
        ),
    )
    def test_exp_to_df_row_ordering(self, _) -> None:
        """
        This test verifies that the returned data frame indexes are
        in the same order as trial index. It mocks _merge_results_if_no_duplicates
        to verify just the ordering of items in the final data frame.
        """
        exp = get_branin_experiment(with_trial=True)
        df = exp_to_df(exp)
        # Check that all 3 rows are in order
        self.assertEqual(len(df), 3)
        for idx, row in df.iterrows():
            self.assertEqual(row["trial_index"], idx)

    @patch(
        "ax.service.utils.report_utils._merge_results_if_no_duplicates",
        autospec=True,
        return_value=pd.DataFrame(
            [
                # Trial indexes are out-of-order.
                {
                    "col1": 1,
                    "arm_name": "a",
                    "trial_status": "FAILED",
                    "generation_method": "Manual",
                    "trial_index": 1,
                },
                {
                    "col1": 2,
                    "arm_name": "b",
                    "trial_status": "COMPLETED",
                    "generation_method": "BO",
                    "trial_index": 2,
                },
                {
                    "col1": 3,
                    "arm_name": "c",
                    "trial_status": "COMPLETED",
                    "generation_method": "Manual",
                    "trial_index": 0,
                },
            ]
        ),
    )
    def test_exp_to_df_col_ordering(self, _) -> None:
        """
        This test verifies that the returned data frame indexes are
        in the same order as trial index. It mocks _merge_results_if_no_duplicates
        to verify just the ordering of items in the final data frame.
        """
        exp = get_branin_experiment(with_trial=True)
        df = exp_to_df(exp)
        self.assertListEqual(
            list(df.columns),
            ["trial_index", "arm_name", "trial_status", "generation_method", "col1"],
        )

    def test_exp_to_df_max_map_value(self) -> None:
        exp = get_test_map_data_experiment(num_trials=3, num_fetches=5, num_complete=0)

        def compute_maximum_map_values_timestamp(
            experiment: Experiment,
        ) -> Dict[int, float]:
            return compute_maximum_map_values(
                experiment=experiment, map_key="timestamp"
            )

        df = exp_to_df(
            exp=exp,
            additional_fields_callables={  # pyre-ignore
                "timestamp": compute_maximum_map_values_timestamp
            },
        )
        self.assertEqual(df["timestamp"].tolist(), [4.0, 4.0, 4.0])

    def test_exp_to_df_trial_timing(self) -> None:
        # 1. test all have started, none have completed
        exp = get_test_map_data_experiment(num_trials=3, num_fetches=5, num_complete=0)
        df = exp_to_df(
            exp=exp,
            trial_attribute_fields=["time_run_started", "time_completed"],
            always_include_field_columns=True,
        )
        self.assertTrue("time_run_started" in list(df.columns))
        self.assertTrue("time_completed" in list(df.columns))
        # since all trials started, all should have values
        self.assertFalse(any(df["time_run_started"].isnull()))
        # since no trials are complete, all should be None
        self.assertTrue(all(df["time_completed"].isnull()))

        # 2. test some trials not started yet
        exp.trials[0]._time_run_started = None
        df = exp_to_df(
            exp=exp, trial_attribute_fields=["time_run_started", "time_completed"]
        )
        # the first trial should have NaN for rel_time_run_started
        self.assertTrue(df["time_run_started"].isnull().iloc[0])

        # 3. test all trials not started yet
        for t in exp.trials.values():
            t._time_run_started = None
        df = exp_to_df(
            exp=exp,
            trial_attribute_fields=["time_run_started", "time_completed"],
            always_include_field_columns=True,
        )
        self.assertTrue(all(df["time_run_started"].isnull()))

        # 4. test some trials are completed
        exp = get_test_map_data_experiment(num_trials=3, num_fetches=5, num_complete=2)
        df = exp_to_df(
            exp=exp, trial_attribute_fields=["time_run_started", "time_completed"]
        )
        # the last trial should have NaN for rel_time_completed
        self.assertTrue(df["time_completed"].isnull().iloc[2])

    def test_exp_to_df_with_failure(self) -> None:
        fail_reason = "test reason"

        # Set up experiment with a failed trial
        exp = get_branin_experiment(with_trial=True)
        exp.trials[0].run()
        exp.trials[0].mark_failed(reason=fail_reason)

        df = exp_to_df(exp)
        self.assertEqual(
            set(EXPECTED_COLUMNS + ["reason"]) - set(df.columns), {OBJECTIVE_NAME}
        )
        self.assertEqual(f"{fail_reason}...", df["reason"].iloc[0])

    def test_exp_to_df(self) -> None:
        # MultiTypeExperiment should fail
        exp = get_multi_type_experiment()
        with self.assertRaisesRegex(ValueError, "MultiTypeExperiment"):
            exp_to_df(exp=exp)

        # exp with no trials should return empty results
        exp = get_branin_experiment()
        df = exp_to_df(exp=exp)
        self.assertEqual(len(df), 0)

        # set up experiment
        exp = get_branin_experiment(with_batch=True)

        # check that pre-run experiment returns all columns except objective
        df = exp_to_df(exp)
        self.assertEqual(set(EXPECTED_COLUMNS) - set(df.columns), {OBJECTIVE_NAME})
        self.assertEqual(len(df.index), len(exp.arms_by_name))

        exp.trials[0].run()
        exp.fetch_data()

        # assert result is df with expected columns and length
        df = exp_to_df(exp=exp)
        self.assertIsInstance(df, pd.DataFrame)
        self.assertListEqual(sorted(df.columns), sorted(EXPECTED_COLUMNS))
        self.assertEqual(len(df.index), len(exp.arms_by_name))

        # test with run_metadata_fields and trial_properties_fields not empty
        # add source to properties
        for _, trial in exp.trials.items():
            trial._properties["source"] = DUMMY_SOURCE
        df = exp_to_df(
            exp, run_metadata_fields=["name"], trial_properties_fields=["source"]
        )
        self.assertIn("name", df.columns)
        self.assertIn("trial_properties_source", df.columns)

        # test column values or types
        self.assertTrue(all(x == 0 for x in df.trial_index))
        self.assertTrue(all(x == "RUNNING" for x in df.trial_status))
        self.assertTrue(all(x == "Sobol" for x in df.generation_method))
        self.assertTrue(all(x == DUMMY_SOURCE for x in df.trial_properties_source))
        self.assertTrue(all(x == "branin_test_experiment_0" for x in df.name))
        for float_column in FLOAT_COLUMNS:
            self.assertTrue(all(isinstance(x, float) for x in df[float_column]))

        # works correctly for failed trials (will need to mock)
        dummy_struct = namedtuple("dummy_struct", "df")
        mock_results = dummy_struct(
            df=pd.DataFrame(
                {
                    "arm_name": ["0_0", "1_0"],
                    "metric_name": [OBJECTIVE_NAME] * 2,
                    "mean": [DUMMY_OBJECTIVE_MEAN] * 2,
                    "sem": [0] * 2,
                    "trial_index": [0, 1],
                    "n": [123] * 2,
                    "frac_nonnull": [1] * 2,
                }
            )
        )
        mock_results.df.index = range(len(exp.trials) * 2, len(exp.trials) * 2 + 2)
        with patch.object(Experiment, "lookup_data", lambda self: mock_results):
            df = exp_to_df(exp=exp)
        # all but two rows should have a metric value of NaN
        # pyre-fixme[16]: `bool` has no attribute `sum`.
        self.assertEqual(pd.isna(df[OBJECTIVE_NAME]).sum(), len(df.index) - 2)

        # an experiment with more results than arms raises an error
        with patch.object(
            Experiment, "lookup_data", lambda self: mock_results
        ), self.assertRaisesRegex(ValueError, "inconsistent experimental state"):
            exp_to_df(exp=get_branin_experiment())

        # custom added trial has a generation_method of Manual
        custom_arm = Arm(name="custom", parameters={"x1": 0, "x2": 0})
        exp.new_trial().add_arm(custom_arm)
        df = exp_to_df(exp)
        self.assertEqual(
            df[df.arm_name == "custom"].iloc[0].generation_method, "Manual"
        )
        # failing feasibility calculation doesn't warns and suppresses error
        observations = [[1.0, 2.0, 3.0], [4.0, 5.0, -6.0], [7.0, 8.0, 9.0]]
        exp = get_experiment_with_observations(
            observations=observations,
            constrained=True,
        )
        with patch(
            f"{exp_to_df.__module__}._is_row_feasible", side_effect=KeyError(DUMMY_MSG)
        ), self.assertLogs(logger="ax", level=WARN) as log:
            exp_to_df(exp)
            self.assertIn(
                f"Feasibility calculation failed with error: '{DUMMY_MSG}'",
                log.output[0],
            )

        # infeasible arm has `is_feasible = False`.
        df = exp_to_df(exp)
        self.assertListEqual(list(df[FEASIBLE_COL_NAME]), [True, False, True])

        # all rows infeasible.
        observations = [[1.0, 2.0, -3.0], [4.0, 5.0, -6.0], [7.0, 8.0, -9.0]]
        exp = get_experiment_with_observations(
            observations=observations,
            constrained=True,
        )
        df = exp_to_df(exp)
        self.assertListEqual(list(df[FEASIBLE_COL_NAME]), [False, False, False])

    def test_get_shortest_unique_suffix_dict(self) -> None:
        expected_output = {
            "abc.123": "abc.123",
            "asdf.abc.123": "asdf.abc.123",
            "def.123": "def.123",
            "abc.456": "456",
            "": "",
            "no_delimiter": "no_delimiter",
        }
        actual_output = _get_shortest_unique_suffix_dict(
            ["abc.123", "abc.456", "def.123", "asdf.abc.123", "", "no_delimiter"]
        )
        self.assertDictEqual(expected_output, actual_output)

    @fast_botorch_optimize
    def test_get_standard_plots(self) -> None:
        exp = get_branin_experiment()
        self.assertEqual(
            len(
                get_standard_plots(
                    experiment=exp, model=get_generation_strategy().model
                )
            ),
            0,
        )
        exp = get_branin_experiment(with_batch=True, minimize=True)
        exp.trials[0].run()
        exp.trials[0].mark_completed()
        model = Models.BOTORCH_MODULAR(experiment=exp, data=exp.fetch_data())
        for gsa, true_objective_metric_name in itertools.product(
            [False, True], ["branin", None]
        ):
            with self.subTest(global_sensitivity_analysis=gsa):
                plots = get_standard_plots(
                    experiment=exp,
                    model=model,
                    global_sensitivity_analysis=gsa,
                    true_objective_metric_name=true_objective_metric_name,
                )
                self.assertEqual(len(plots), 8 if true_objective_metric_name else 6)
                self.assertTrue(all(isinstance(plot, go.Figure) for plot in plots))

        # Raise an exception in one plot and make sure we generate the others
        for plot_function, num_expected_plots in [
            [_get_curve_plot_dropdown, 8],  # Not used
            [_get_objective_trace_plot, 6],
            [_objective_vs_true_objective_scatter, 7],
            [_get_objective_v_param_plots, 6],
            [_get_cross_validation_plots, 7],
            [plot_feature_importance_by_feature_plotly, 6],
        ]:
            with mock.patch(
                # pyre-ignore
                f"ax.service.utils.report_utils.{plot_function.__name__}",
                side_effect=Exception(),
            ):
                plots = get_standard_plots(
                    experiment=exp,
                    model=model,
                    global_sensitivity_analysis=True,
                    true_objective_metric_name="branin",
                )
            self.assertEqual(len(plots), num_expected_plots)
            self.assertTrue(all(isinstance(plot, go.Figure) for plot in plots))

    @fast_botorch_optimize
    def test_get_standard_plots_moo(self) -> None:
        exp = get_branin_experiment_with_multi_objective(with_batch=True)
        exp.optimization_config.objective.objectives[0].minimize = False
        exp.optimization_config.objective.objectives[1].minimize = True
        checked_cast(
            MultiObjectiveOptimizationConfig, exp.optimization_config
        )._objective_thresholds = [
            ObjectiveThreshold(
                metric=exp.metrics["branin_a"], op=ComparisonOp.GEQ, bound=-100.0
            ),
            ObjectiveThreshold(
                metric=exp.metrics["branin_b"], op=ComparisonOp.LEQ, bound=100.0
            ),
        ]
        exp.trials[0].run()
        # NOTE: level set to INFO in this block, because the global sensitivity
        # analysis raises an INFO level log entry here. Leaving level=WARN here
        # actually passes on Python 3.8 because of a language internal bug. See
        # https://bugs.python.org/issue41943 for more information.
        with self.assertLogs(logger="ax", level=INFO) as log:
            plots = get_standard_plots(
                experiment=exp, model=Models.MOO(experiment=exp, data=exp.fetch_data())
            )
            self.assertEqual(len(log.output), 5)
            self.assertIn(
                "Pareto plotting not supported for experiments with relative objective "
                "thresholds.",
                log.output[0],
            )
            self.assertIn(
                "Failed to compute signed global feature sensitivities",
                log.output[1],
            )
            self.assertIn(
                "Failed to compute unsigned feature sensitivities:",
                log.output[2],
            )
            created_plots_logs = set(log.output[2:])
            for metric_suffix in ("a", "b"):
                expected_msg = (
                    "Created contour plots for metric branin_"
                    f"{metric_suffix} and parameters ['x2', 'x1']"
                )
                self.assertTrue(any(expected_msg in msg for msg in created_plots_logs))
        self.assertEqual(len(plots), 6)

    @fast_botorch_optimize
    def test_get_standard_plots_moo_relative_constraints(self) -> None:
        exp = get_branin_experiment_with_multi_objective(with_batch=True)
        exp.optimization_config.objective.objectives[0].minimize = False
        exp.optimization_config.objective.objectives[1].minimize = True
        checked_cast(
            MultiObjectiveOptimizationConfig, exp.optimization_config
        )._objective_thresholds = [
            ObjectiveThreshold(
                metric=exp.metrics["branin_a"], op=ComparisonOp.GEQ, bound=-100.0
            ),
            ObjectiveThreshold(
                metric=exp.metrics["branin_b"], op=ComparisonOp.LEQ, bound=100.0
            ),
        ]
        exp.trials[0].run()

        for ot in checked_cast(
            MultiObjectiveOptimizationConfig, exp.optimization_config
        )._objective_thresholds:
            ot.relative = False
        plots = get_standard_plots(
            experiment=exp, model=Models.MOO(experiment=exp, data=exp.fetch_data())
        )
        self.assertEqual(len(plots), 8)

    @fast_botorch_optimize
    def test_get_standard_plots_moo_no_objective_thresholds(self) -> None:
        exp = get_branin_experiment_with_multi_objective(with_batch=True)
        exp.optimization_config.objective.objectives[0].minimize = False
        exp.optimization_config.objective.objectives[1].minimize = True
        exp.trials[0].run()
        plots = get_standard_plots(
            experiment=exp, model=Models.MOO(experiment=exp, data=exp.fetch_data())
        )
        self.assertEqual(len(plots), 8)

    @fast_botorch_optimize
    def test_get_standard_plots_map_data(self) -> None:
        exp = get_branin_experiment_with_timestamp_map_metric(with_status_quo=True)
        exp.new_trial().add_arm(exp.status_quo)
        exp.trials[0].run()
        exp.new_trial(
            generator_run=Models.SOBOL(search_space=exp.search_space).gen(n=1)
        )
        exp.trials[1].run()
        for t in exp.trials.values():
            t.mark_completed()
        plots = get_standard_plots(
            experiment=exp,
            model=Models.BOTORCH_MODULAR(experiment=exp, data=exp.fetch_data()),
            true_objective_metric_name="branin",
        )

        self.assertEqual(len(plots), 10)
        self.assertTrue(all(isinstance(plot, go.Figure) for plot in plots))
        self.assertIn(
            "Objective branin_map vs. True Objective Metric branin",
            [p.layout.title.text for p in plots],  # pyre-ignore[16]
        )

        with self.assertRaisesRegex(
            ValueError, "Please add a valid true_objective_metric_name"
        ):
            plots = get_standard_plots(
                experiment=exp,
                model=Models.BOTORCH_MODULAR(experiment=exp, data=exp.fetch_data()),
                true_objective_metric_name="not_present",
            )

    @fast_botorch_optimize
    def test_skip_contour_high_dimensional(self) -> None:
        exp = get_high_dimensional_branin_experiment()
        # Initial Sobol points
        sobol = Models.SOBOL(search_space=exp.search_space)
        for _ in range(1):
            exp.new_trial(sobol.gen(1)).run()
        model = Models.GPEI(
            experiment=exp,
            data=exp.fetch_data(),
        )
        with self.assertLogs(logger="ax", level=INFO) as log:
            _get_objective_v_param_plots(
                experiment=exp, model=model, max_num_contour_plots=2
            )
            self.assertEqual(len(log.output), 1)
            self.assertIn(
                "Created contour plots for metric objective and parameters",
                log.output[0],
            )
        with self.assertLogs(logger="ax", level=WARN) as log:
            _get_objective_v_param_plots(
                experiment=exp, model=model, max_num_contour_plots=1
            )
            self.assertEqual(len(log.output), 1)
            self.assertIn(
                "Skipping creation of contour plots",
                log.output[0],
            )
        with self.assertLogs(logger="ax", level=WARN) as log:
            _get_objective_v_param_plots(
                experiment=exp,
                model=model,
                max_num_contour_plots=1,
                max_num_slice_plots=10,
            )
            # Creates two warnings, one for slice plots and one for contour plots.
            self.assertEqual(len(log.output), 2)

    def test_get_metric_name_pairs(self) -> None:
        exp = get_branin_experiment(with_trial=True)
        exp._optimization_config = MultiObjectiveOptimizationConfig(
            objective=MultiObjective(
                objectives=[
                    Objective(metric=Metric("m0"), minimize=False),
                    Objective(metric=Metric("m1"), minimize=False),
                    Objective(metric=Metric("m2"), minimize=False),
                    Objective(metric=Metric("m3"), minimize=False),
                    Objective(metric=Metric("m4"), minimize=False),
                ]
            )
        )
        with self.assertLogs(logger="ax", level=WARN) as log:
            metric_name_pairs = _get_metric_name_pairs(experiment=exp)
            self.assertEqual(len(log.output), 1)
            self.assertIn(
                "Creating pairwise Pareto plots for the first `use_n_metrics",
                log.output[0],
            )
        self.assertListEqual(
            list(metric_name_pairs),
            list(itertools.combinations([f"m{i}" for i in range(4)], 2)),
        )

    def test_compare_to_baseline(self) -> None:
        """Test that compare to baseline parses arm df properly,
        obtains the objective metric values based
        on the provided OptimizationConfig, and
        produces the intended text
        """
        self.maxDiff = None
        OBJECTIVE_METRIC = "foo"

        data = [
            {"trial_index": 0, "arm_name": BASELINE_ARM_NAME, OBJECTIVE_METRIC: 0.2},
            {"trial_index": 1, "arm_name": "dummy", OBJECTIVE_METRIC: 0.5},
            {"trial_index": 2, "arm_name": "optimal", OBJECTIVE_METRIC: 2.5},
            {"trial_index": 3, "arm_name": "bad_optimal", OBJECTIVE_METRIC: 0.05},
        ]
        arms_df = pd.DataFrame(data)

        with patch(
            "ax.service.utils.report_utils.exp_to_df",
            return_value=arms_df,
        ):
            true_obj_metric = Metric(name=OBJECTIVE_METRIC, lower_is_better=False)
            experiment = Experiment(
                search_space=get_branin_search_space(),
                tracking_metrics=[true_obj_metric],
            )

            optimization_config = OptimizationConfig(
                objective=Objective(metric=true_obj_metric, minimize=False),
                outcome_constraints=[],
            )
            experiment.optimization_config = optimization_config

            comparison_arm_names = ["optimal"]

            result = compare_to_baseline(
                experiment=experiment,
                optimization_config=None,
                comparison_arm_names=comparison_arm_names,
                baseline_arm_name=BASELINE_ARM_NAME,
            )

            output_text = _format_comparison_string(
                comparison_arm_name=comparison_arm_names[0],
                baseline_arm_name=BASELINE_ARM_NAME,
                objective_name=OBJECTIVE_METRIC,
                percent_change=1150.0,
                baseline_value=0.2,
                comparison_value=2.5,
                digits=2,
            )

            self.assertNotEqual(result, None)
            self.assertEqual(result, output_text)

            bad_comparison_arm_names = ["bad_optimal"]
            # because result increased from baseline, no improvement result returned
            bad_result = compare_to_baseline(
                experiment=experiment,
                optimization_config=None,
                comparison_arm_names=bad_comparison_arm_names,
                baseline_arm_name=BASELINE_ARM_NAME,
            )
            self.assertEqual(bad_result, None)

    def test_compare_to_baseline_pass_in_opt(self) -> None:
        """Test that compare to baseline parses arm df properly,
        obtains the objective metric values based
        on the provided OptimizationConfig, and
        produces the intended text
        """
        self.maxDiff = None
        OBJECTIVE_METRIC = "foo"

        data = [
            {"trial_index": 0, "arm_name": BASELINE_ARM_NAME, OBJECTIVE_METRIC: 0.2},
            {"trial_index": 1, "arm_name": "dummy", OBJECTIVE_METRIC: 0.5},
            {"trial_index": 2, "arm_name": "optimal", OBJECTIVE_METRIC: 2.5},
            {"trial_index": 3, "arm_name": "bad_optimal", OBJECTIVE_METRIC: 0.05},
        ]
        arms_df = pd.DataFrame(data)

        with patch(
            "ax.service.utils.report_utils.exp_to_df",
            return_value=arms_df,
        ):
            true_obj_metric = Metric(name=OBJECTIVE_METRIC, lower_is_better=False)
            experiment = Experiment(
                search_space=get_branin_search_space(),
                tracking_metrics=[true_obj_metric],
                optimization_config=None,
            )

            optimization_config = OptimizationConfig(
                objective=Objective(metric=true_obj_metric, minimize=False),
                outcome_constraints=[],
            )
            experiment.optimization_config = optimization_config

            comparison_arm_names = ["optimal"]

            result = compare_to_baseline(
                experiment=experiment,
                optimization_config=optimization_config,
                comparison_arm_names=comparison_arm_names,
                baseline_arm_name=BASELINE_ARM_NAME,
            )

            output_text = _format_comparison_string(
                comparison_arm_name=comparison_arm_names[0],
                baseline_arm_name=BASELINE_ARM_NAME,
                objective_name=OBJECTIVE_METRIC,
                percent_change=1150.0,
                baseline_value=0.2,
                comparison_value=2.5,
                digits=2,
            )

            self.assertNotEqual(result, None)
            self.assertEqual(result, output_text)

    def test_compare_to_baseline_minimize(self) -> None:
        """Test that compare to baseline parses arm df properly,
        obtains the objective metric values based
        on the provided OptimizationConfig, and
        produces the intended text.
        For the minimize case.
        Also will use a custom baseline arm name.
        """
        self.maxDiff = None
        OBJECTIVE_METRIC = "foo"
        custom_baseline_arm_name = "custom_baseline"

        data = [
            {
                "trial_index": 0,
                "arm_name": custom_baseline_arm_name,
                OBJECTIVE_METRIC: 0.2,
            },
            {"trial_index": 1, "arm_name": "dummy", OBJECTIVE_METRIC: 0.5},
            {"trial_index": 2, "arm_name": "optimal", OBJECTIVE_METRIC: 0.1},
            {"trial_index": 3, "arm_name": "bad_optimal", OBJECTIVE_METRIC: 1.0},
        ]
        arms_df = pd.DataFrame(data)

        with patch(
            "ax.service.utils.report_utils.exp_to_df",
            return_value=arms_df,
        ):
            true_obj_metric = Metric(name=OBJECTIVE_METRIC, lower_is_better=True)
            experiment = Experiment(
                search_space=get_branin_search_space(),
                tracking_metrics=[true_obj_metric],
            )

            optimization_config = OptimizationConfig(
                objective=Objective(metric=true_obj_metric, minimize=True),
                outcome_constraints=[],
            )
            experiment.optimization_config = optimization_config

            comparison_arm_names = ["optimal"]

            result = compare_to_baseline(
                experiment=experiment,
                optimization_config=None,
                comparison_arm_names=comparison_arm_names,
                baseline_arm_name=custom_baseline_arm_name,
            )

            output_text = _format_comparison_string(
                comparison_arm_name=comparison_arm_names[0],
                baseline_arm_name=custom_baseline_arm_name,
                objective_name=OBJECTIVE_METRIC,
                percent_change=50.0,
                baseline_value=0.2,
                comparison_value=0.1,
                digits=2,
            )

            self.assertNotEqual(result, None)
            self.assertEqual(result, output_text)

            bad_comparison_arm_names = ["bad_optimal"]
            # because result increased from baseline, no improvement result returned
            bad_result = compare_to_baseline(
                experiment=experiment,
                optimization_config=None,
                comparison_arm_names=bad_comparison_arm_names,
                baseline_arm_name=BASELINE_ARM_NAME,
            )
            self.assertEqual(bad_result, None)

    def test_compare_to_baseline_edge_case(self) -> None:
        """Test that compare to baseline parses arm df properly,
        obtains the objective metric values based
        on the provided OptimizationConfig, and
        produces the intended text
        """
        self.maxDiff = None
        OBJECTIVE_METRIC = "foo"

        true_obj_metric = Metric(name=OBJECTIVE_METRIC, lower_is_better=True)
        experiment = Experiment(
            search_space=get_branin_search_space(),
            tracking_metrics=[true_obj_metric],
        )

        optimization_config = OptimizationConfig(
            objective=Objective(metric=true_obj_metric, minimize=True),
            outcome_constraints=[],
        )
        experiment.optimization_config = optimization_config
        comparison_arm_names = ["optimal"]

        # baseline value is 0
        data = [
            {"trial_index": 0, "arm_name": BASELINE_ARM_NAME, OBJECTIVE_METRIC: 0.0},
            {"trial_index": 1, "arm_name": "optimal", OBJECTIVE_METRIC: 1.0},
        ]
        arms_df = pd.DataFrame(data)

        with patch(
            "ax.service.utils.report_utils.exp_to_df",
            return_value=arms_df,
        ):
            with self.assertLogs("ax", level=INFO) as log:
                self.assertEqual(
                    compare_to_baseline(
                        experiment=experiment,
                        optimization_config=None,
                        comparison_arm_names=comparison_arm_names,
                        baseline_arm_name=BASELINE_ARM_NAME,
                    ),
                    None,
                )
                self.assertTrue(
                    any(
                        (
                            "compare_to_baseline: baseline has value of 0"
                            ", can't compute percent change."
                        )
                        in log_str
                        for log_str in log.output
                    ),
                    log.output,
                )

        # no best arm names
        with patch(
            "ax.service.utils.report_utils.exp_to_df",
            return_value=arms_df,
        ):
            with self.assertLogs("ax", level=INFO) as log:
                self.assertEqual(
                    compare_to_baseline(
                        experiment=experiment,
                        optimization_config=None,
                        comparison_arm_names=None,
                        baseline_arm_name=BASELINE_ARM_NAME,
                    ),
                    None,
                )
                self.assertTrue(
                    any(
                        (
                            "compare_to_baseline: comparison_arm_names not provided."
                            " Returning None."
                        )
                        in log_str
                        for log_str in log.output
                    ),
                    log.output,
                )

        # no optimization config
        with patch(
            "ax.service.utils.report_utils.exp_to_df",
            return_value=arms_df,
        ):
            with self.assertLogs("ax", level=INFO) as log:
                exp_no_opt = Experiment(
                    search_space=get_branin_search_space(),
                    tracking_metrics=[true_obj_metric],
                    optimization_config=None,
                )
                self.assertEqual(
                    compare_to_baseline(
                        experiment=exp_no_opt,
                        optimization_config=None,
                        comparison_arm_names=comparison_arm_names,
                        baseline_arm_name=BASELINE_ARM_NAME,
                    ),
                    None,
                )
                self.assertEqual(exp_no_opt.optimization_config, None)
                self.assertTrue(
                    any(
                        (
                            "compare_to_baseline: optimization_config neither provided "
                            "in inputs nor present on experiment."
                        )
                        in log_str
                        for log_str in log.output
                    ),
                    log.output,
                )

    def test_compare_to_baseline_arms_not_found(self) -> None:
        self.maxDiff = None
        OBJECTIVE_METRIC = "foo"

        true_obj_metric = Metric(name=OBJECTIVE_METRIC, lower_is_better=True)
        experiment = Experiment(
            search_space=get_branin_search_space(),
            tracking_metrics=[true_obj_metric],
        )

        optimization_config = OptimizationConfig(
            objective=Objective(metric=true_obj_metric, minimize=True),
            outcome_constraints=[],
        )
        experiment.optimization_config = optimization_config
        comparison_arm_names = ["optimal"]

        # baseline value is 0
        data = [
            {"trial_index": 0, "arm_name": BASELINE_ARM_NAME, OBJECTIVE_METRIC: 0.0},
            {"trial_index": 1, "arm_name": "optimal", OBJECTIVE_METRIC: 1.0},
        ]
        arms_df = pd.DataFrame(data)

        # no arms df
        with patch(
            "ax.service.utils.report_utils.exp_to_df",
            return_value=None,
        ):
            with self.assertLogs("ax", level=INFO) as log:
                self.assertEqual(
                    compare_to_baseline(
                        experiment=experiment,
                        optimization_config=None,
                        comparison_arm_names=comparison_arm_names,
                        baseline_arm_name=BASELINE_ARM_NAME,
                    ),
                    None,
                )
                self.assertTrue(
                    any(
                        ("compare_to_baseline: arms_df is None.") in log_str
                        for log_str in log.output
                    ),
                    log.output,
                )

        # best arms df is none
        with patch(
            "ax.service.utils.report_utils.exp_to_df",
            return_value=arms_df,
        ):
            with self.assertLogs("ax", level=INFO) as log:
                comparison_arm_not_found = ["unknown_arm"]
                self.assertEqual(
                    compare_to_baseline(
                        experiment=experiment,
                        optimization_config=None,
                        comparison_arm_names=comparison_arm_not_found,
                        baseline_arm_name=BASELINE_ARM_NAME,
                    ),
                    None,
                )
                self.assertTrue(
                    any(
                        ("compare_to_baseline: comparison_arm_df has no rows.")
                        in log_str
                        for log_str in log.output
                    ),
                    log.output,
                )

        # baseline not found in arms_df
        with patch(
            "ax.service.utils.report_utils.exp_to_df",
            return_value=arms_df,
        ):
            experiment_with_status_quo = experiment
            experiment_with_status_quo.status_quo = Arm(
                name=BASELINE_ARM_NAME,
                parameters={"x1": 0, "x2": 0},
            )
            baseline_arm_name = "not_baseline_arm_in_dataframe"
            with self.assertLogs("ax", level=INFO) as log:
                self.assertEqual(
                    compare_to_baseline(
                        experiment=experiment_with_status_quo,
                        optimization_config=None,
                        comparison_arm_names=comparison_arm_names,
                        baseline_arm_name=baseline_arm_name,
                    ),
                    None,
                )
                self.assertTrue(
                    any(
                        (
                            f"compare_to_baseline: baseline row: {baseline_arm_name=}"
                            " not found in arms"
                        )
                        in log_str
                        for log_str in log.output
                    ),
                    log.output,
                )

    def test_compare_to_baseline_moo(self) -> None:
        """Test that compare to baseline errors out correctly
        for multi objective problems

        """
        self.maxDiff = None

        data = [
            {
                "trial_index": 0,
                "arm_name": BASELINE_ARM_NAME,
                "m0": 1.0,
                "m1": 1.0,
                "m2": 1.0,
                "m3": 1.0,
            },
            {
                "trial_index": 1,
                "arm_name": "opt_0",
                "m0": 2.5,
                "m1": 0.2,
                "m2": 1.0,
                "m3": 1.0,
            },
            {
                "trial_index": 2,
                "arm_name": "opt_1_min",
                "m0": 1.2,
                "m1": -0.1,
                "m2": 1.0,
                "m3": 1.0,
            },
            {
                "trial_index": 3,
                "arm_name": "opt_3",
                "m0": 0.5,
                "m1": 0.5,
                "m2": 1.0,
                "m3": 1.01,
            },
        ]
        arms_df = pd.DataFrame(data)

        with patch(
            "ax.service.utils.report_utils.exp_to_df",
            return_value=arms_df,
        ):
            m0 = Metric(name="m0", lower_is_better=False)
            m1 = Metric(name="m1", lower_is_better=True)
            m2 = Metric(name="m2", lower_is_better=False)
            m3 = Metric(name="m3", lower_is_better=False)
            experiment = Experiment(
                search_space=get_branin_search_space(),
                tracking_metrics=[m0, m1, m2, m3],
            )

            optimization_config = MultiObjectiveOptimizationConfig(
                objective=MultiObjective(
                    objectives=[
                        Objective(metric=Metric("m0"), minimize=False),
                        Objective(metric=Metric("m1"), minimize=True),
                        Objective(metric=Metric("m3"), minimize=False),
                    ]
                )
            )
            experiment.optimization_config = optimization_config
            self.assertEqual(True, experiment.is_moo_problem)

            comparison_arm_names = ["opt_0", "opt_1_min", "opt_3"]

            preamble = (
                "Below is the greatest improvement, if any,"
                " achieved for each objective metric \n"
            )
            output_text_0 = _format_comparison_string(
                comparison_arm_name="opt_0",
                baseline_arm_name=BASELINE_ARM_NAME,
                objective_name="m0",
                percent_change=150.0,
                baseline_value=1.0,
                comparison_value=2.5,
                digits=2,
            )
            output_text_1 = _format_comparison_string(
                comparison_arm_name="opt_1_min",
                baseline_arm_name=BASELINE_ARM_NAME,
                objective_name="m1",
                percent_change=110.0,
                baseline_value=1.0,
                comparison_value=-0.1,
                digits=2,
            )
            output_text_3 = _format_comparison_string(
                comparison_arm_name="opt_3",
                baseline_arm_name=BASELINE_ARM_NAME,
                objective_name="m3",
                percent_change=1.0,
                baseline_value=1.0,
                comparison_value=1.01,
                digits=2,
            )

            result = not_none(
                compare_to_baseline(
                    experiment=experiment,
                    optimization_config=None,
                    comparison_arm_names=comparison_arm_names,
                    baseline_arm_name=BASELINE_ARM_NAME,
                ),
            )

            expected_result = (
                preamble
                + " \n* "
                + output_text_0
                + " \n* "
                + output_text_1
                + " \n* "
                + output_text_3
            )

            self.assertEqual(result, expected_result)

    def test_compare_to_baseline_equal(self) -> None:
        """Test case where baseline value is equal to optimal value"""
        self.maxDiff = None
        OBJECTIVE_METRIC = "foo"
        custom_baseline_arm_name = "custom_baseline"

        data = [
            {
                "trial_index": 0,
                "arm_name": custom_baseline_arm_name,
                OBJECTIVE_METRIC: 0.2,
            },
            {"trial_index": 1, "arm_name": "dummy", OBJECTIVE_METRIC: 0.5},
            {"trial_index": 2, "arm_name": "optimal", OBJECTIVE_METRIC: 0.5},
            {"trial_index": 3, "arm_name": "equal", OBJECTIVE_METRIC: 0.2},
        ]
        arms_df = pd.DataFrame(data)

        true_obj_metric = Metric(name=OBJECTIVE_METRIC, lower_is_better=True)
        experiment = Experiment(
            search_space=get_branin_search_space(),
            tracking_metrics=[true_obj_metric],
        )

        optimization_config = OptimizationConfig(
            objective=Objective(metric=true_obj_metric, minimize=True),
            outcome_constraints=[],
        )
        experiment.optimization_config = optimization_config

        comparison_arm_names = ["equal"]

        with patch(
            "ax.service.utils.report_utils.exp_to_df",
            return_value=arms_df,
        ):
            result = compare_to_baseline(
                experiment=experiment,
                optimization_config=optimization_config,
                comparison_arm_names=comparison_arm_names,
                baseline_arm_name=custom_baseline_arm_name,
            )

            self.assertIsNone(result)

    def test_compare_to_baseline_select_baseline_arm(self) -> None:
        OBJECTIVE_METRIC = "objective"
        true_obj_metric = Metric(name=OBJECTIVE_METRIC, lower_is_better=True)
        experiment = Experiment(
            search_space=get_branin_search_space(),
            tracking_metrics=[true_obj_metric],
        )

        # specified baseline
        data = [
            {
                "trial_index": 0,
                "arm_name": "m_0",
                OBJECTIVE_METRIC: 0.2,
            },
            {
                "trial_index": 1,
                "arm_name": BASELINE_ARM_NAME,
                OBJECTIVE_METRIC: 0.2,
            },
            {
                "trial_index": 2,
                "arm_name": "status_quo",
                OBJECTIVE_METRIC: 0.2,
            },
        ]
        arms_df = pd.DataFrame(data)
        self.assertEqual(
            select_baseline_arm(
                experiment=experiment,
                arms_df=arms_df,
                baseline_arm_name=BASELINE_ARM_NAME,
            ),
            (BASELINE_ARM_NAME, False),
        )

        # specified baseline arm not in trial
        wrong_baseline_name = "wrong_baseline_name"
        with self.assertRaisesRegex(
            ValueError,
            "compare_to_baseline: baseline row: .*" + " not found in arms",
        ):
            select_baseline_arm(
                experiment=experiment,
                arms_df=arms_df,
                baseline_arm_name=wrong_baseline_name,
            ),

        # status quo baseline arm
        experiment_with_status_quo = copy.deepcopy(experiment)
        experiment_with_status_quo.status_quo = Arm(
            name="status_quo",
            parameters={"x1": 0, "x2": 0},
        )
        self.assertEqual(
            select_baseline_arm(
                experiment=experiment_with_status_quo,
                arms_df=arms_df,
                baseline_arm_name=None,
            ),
            ("status_quo", False),
        )
        # first arm from trials
        custom_arm = Arm(name="m_0", parameters={"x1": 0.1, "x2": 0.2})
        experiment.new_trial().add_arm(custom_arm)
        self.assertEqual(
            select_baseline_arm(
                experiment=experiment,
                arms_df=arms_df,
                baseline_arm_name=None,
            ),
            ("m_0", True),
        )

        # none selected
        experiment_with_no_valid_baseline = Experiment(
            search_space=get_branin_search_space(),
            tracking_metrics=[true_obj_metric],
        )
        experiment_with_no_valid_baseline.status_quo = Arm(
            name="not found",
            parameters={"x1": 0, "x2": 0},
        )
        custom_arm = Arm(name="also not found", parameters={"x1": 0.1, "x2": 0.2})
        experiment_with_no_valid_baseline.new_trial().add_arm(custom_arm)
        with self.assertRaisesRegex(
            ValueError, "compare_to_baseline: could not find valid baseline arm"
        ):
            select_baseline_arm(
                experiment=experiment_with_no_valid_baseline,
                arms_df=arms_df,
                baseline_arm_name=None,
            )

    def test_warn_if_unpredictable_metrics(self) -> None:
        expected_msg = (
            "The following metric(s) are behaving unpredictably and may be noisy or "
            "misconfigured: ['branin']. Please check that they are measuring the "
            "intended quantity, and are expected to vary reliably as a function of "
            "your parameters."
        )

        # Create scheduler and run a few trials.
        exp = get_branin_experiment()
        gs = GenerationStrategy(
            steps=[
                GenerationStep(
                    model=Models.SOBOL,
                    num_trials=3,
                    min_trials_observed=3,
                    max_parallelism=3,
                ),
                GenerationStep(model=Models.GPEI, num_trials=-1, max_parallelism=3),
            ]
        )
        gs.experiment = exp
        scheduler = Scheduler(
            generation_strategy=gs, experiment=exp, options=SchedulerOptions()
        )
        scheduler.run_n_trials(1)
        msg = warn_if_unpredictable_metrics(
            experiment=exp,
            generation_strategy=gs,
            model_fit_threshold=1.0,
        )
        self.assertIsNone(msg)

        scheduler.run_n_trials(3)

        # Set fitted model to None to test refitting.
        scheduler.generation_strategy._curr.model_spec_to_gen_from._fitted_model = None

        # Threshold 1.0 (should always generate a warning)
        msg = warn_if_unpredictable_metrics(
            experiment=exp,
            generation_strategy=gs,
            model_fit_threshold=1.0,
        )
        self.assertEqual(msg, expected_msg)

        # Threshold -1.0 (should never generate a warning)
        msg = warn_if_unpredictable_metrics(
            experiment=exp,
            generation_strategy=gs,
            model_fit_threshold=-1.0,
        )
        self.assertIsNone(msg)

        # Test with no optimization config.
        exp._tracking_metrics = exp.metrics
        exp._optimization_config = None
        msg = warn_if_unpredictable_metrics(
            experiment=exp,
            generation_strategy=gs,
            model_fit_threshold=1.0,
        )
        self.assertEqual(msg, expected_msg)

        # Test with manually specified metric_names.
        msg = warn_if_unpredictable_metrics(
            experiment=exp,
            generation_strategy=gs,
            model_fit_threshold=1.0,
            metric_names=["branin"],
        )
        self.assertEqual(msg, expected_msg)

        # Test with metric name that isn't in the experiment.
        with self.assertRaisesRegex(
            ValueError, "Invalid metric names: {'bad_metric_name'}"
        ):
            warn_if_unpredictable_metrics(
                experiment=exp,
                generation_strategy=gs,
                model_fit_threshold=1.0,
                metric_names=["bad_metric_name"],
            )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import os
from glob import glob

from ax.utils.common.testutils import TestCase


class InitTest(TestCase):
    def test_InitFiles(self) -> None:
        """__init__.py files are necessary when not using buck targets"""
        for root, _dirs, files in os.walk("./ax/ax", topdown=False):
            if len(glob(f"{root}/**/*.py", recursive=True)) > 0:
                with self.subTest(root):
                    self.assertTrue(
                        "__init__.py" in files,
                        "directory " + root + " does not contain a .__init__.py file",
                    )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from typing import Dict, Tuple

import numpy as np
from ax.core.types import TParameterization
from ax.exceptions.core import OptimizationShouldStop
from ax.global_stopping.strategies.base import BaseGlobalStoppingStrategy
from ax.service.ax_client import AxClient
from ax.service.utils.instantiation import ObjectiveProperties
from ax.utils.common.testutils import TestCase
from ax.utils.measurement.synthetic_functions import branin
from ax.utils.testing.core_stubs import DummyGlobalStoppingStrategy


class TestGlobalStoppingIntegration(TestCase):
    def get_ax_client_for_branin(
        self,
        global_stopping_strategy: BaseGlobalStoppingStrategy,
    ) -> AxClient:
        """
        Instantiates an AxClient for the branin experiment with the specified
        global stopping strategy.
        """
        ax_client = AxClient(global_stopping_strategy=global_stopping_strategy)
        ax_client.create_experiment(
            name="branin_test_experiment",
            parameters=[
                {
                    "name": "x1",
                    "type": "range",
                    "bounds": [-5.0, 10.0],
                },
                {
                    "name": "x2",
                    "type": "range",
                    "bounds": [0.0, 15.0],
                },
            ],
            objectives={"branin": ObjectiveProperties(minimize=True)},
        )
        return ax_client

    def evaluate(self, parameters: TParameterization) -> Dict[str, Tuple[float, float]]:
        """Evaluates the parameters for branin experiment."""
        x = np.array([parameters.get(f"x{i+1}") for i in range(2)])
        # pyre-fixme[7]: Expected `Dict[str, Tuple[float, float]]` but got
        #  `Dict[str, Tuple[Union[float, ndarray], float]]`.
        return {"branin": (branin(x), 0.0)}

    def test_global_stopping_integration(self) -> None:
        """
        Specifying a dummy global stopping strategy which stops
        the optimization after 3 trials are completed.
        """
        global_stopping_strategy = DummyGlobalStoppingStrategy(
            min_trials=2, trial_to_stop=3
        )
        ax_client = self.get_ax_client_for_branin(
            global_stopping_strategy=global_stopping_strategy
        )

        # Running the first 3 iterations.
        for _ in range(3):
            parameters, trial_index = ax_client.get_next_trial()
            ax_client.complete_trial(
                trial_index=trial_index,
                # pyre-fixme[6]: For 2nd param expected `Union[Dict[str, Union[Tuple[...
                raw_data=self.evaluate(parameters),
            )

        # Trying to run the 4th iteration, which should raise
        exception = OptimizationShouldStop(message="Stop the optimization.")
        with self.assertRaises(OptimizationShouldStop) as cm:
            parameters, trial_index = ax_client.get_next_trial()
        # Assert Exception's message is unchanged.
        self.assertEqual(cm.exception.message, exception.message)

        # Trying to run the 4th iteration by overruling the stopping strategy.
        parameters, trial_index = ax_client.get_next_trial(force=True)
        self.assertIsNotNone(parameters)

        # Test the property & setter.
        self.assertIs(ax_client.global_stopping_strategy, global_stopping_strategy)
        new_gss = DummyGlobalStoppingStrategy(min_trials=5, trial_to_stop=5)
        ax_client.global_stopping_strategy = new_gss
        self.assertIs(ax_client.global_stopping_strategy, new_gss)
        self.assertIs(ax_client._global_stopping_strategy, new_gss)

    def test_min_trials(self) -> None:
        """
        Tests the min_trials mechanism of the stopping strategy; that is,
        the stopping strategy should not take effect before min_trials trials
        are completed.
        """
        global_stopping_strategy = DummyGlobalStoppingStrategy(
            min_trials=3, trial_to_stop=2
        )
        ax_client = self.get_ax_client_for_branin(
            global_stopping_strategy=global_stopping_strategy
        )

        # Running the first 2 iterations.
        for _ in range(2):
            parameters, trial_index = ax_client.get_next_trial()
            ax_client.complete_trial(
                trial_index=trial_index,
                # pyre-fixme[6]: For 2nd param expected `Union[Dict[str, Union[Tuple[...
                raw_data=self.evaluate(parameters),
            )

        # Since min_trials=3, GSS should not stop creating the 3rd iteration.
        parameters, trial_index = ax_client.get_next_trial()
        ax_client.complete_trial(
            trial_index=trial_index,
            # pyre-fixme[6]: For 2nd param expected `Union[Dict[str, Union[Tuple[Unio...
            raw_data=self.evaluate(parameters),
        )
        self.assertIsNotNone(parameters)

        # Now, GSS should stop creating the 4th iteration.
        exception = OptimizationShouldStop(message="Stop the optimization.")
        with self.assertRaises(OptimizationShouldStop) as cm:
            parameters, trial_index = ax_client.get_next_trial()
        # Assert Exception's message is unchanged.
        self.assertEqual(cm.exception.message, exception.message)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from typing import Callable, NamedTuple, Optional

from ax.storage.sqa_store.decoder import Decoder
from ax.storage.sqa_store.encoder import Encoder
from ax.storage.sqa_store.sqa_config import SQAConfig


class DBSettings(NamedTuple):
    """
    Defines behavior for loading/saving experiment to/from db.
    Either creator or url must be specified as a way to connect to the SQL db.
    """

    # pyre-fixme[24]: Generic type `Callable` expects 2 type parameters.
    creator: Optional[Callable] = None
    decoder: Decoder = Decoder(config=SQAConfig())
    encoder: Encoder = Encoder(config=SQAConfig())
    url: Optional[str] = None



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import logging
import warnings
from math import ceil
from typing import Any, cast, Dict, Optional, Type, Union

import torch
from ax.core.experiment import Experiment
from ax.core.optimization_config import OptimizationConfig
from ax.core.parameter import ChoiceParameter, ParameterType, RangeParameter
from ax.core.search_space import SearchSpace
from ax.modelbridge.generation_strategy import GenerationStep, GenerationStrategy
from ax.modelbridge.registry import MODEL_KEY_TO_MODEL_SETUP, ModelRegistryBase, Models
from ax.modelbridge.transforms.base import Transform
from ax.modelbridge.transforms.winsorize import Winsorize
from ax.models.torch.botorch_modular.model import BoTorchModel as ModularBoTorchModel
from ax.models.types import TConfig
from ax.models.winsorization_config import WinsorizationConfig
from ax.utils.common.logger import get_logger
from ax.utils.common.typeutils import not_none


logger: logging.Logger = get_logger(__name__)


DEFAULT_BAYESIAN_PARALLELISM = 3
# `BO_MIXED` optimizes all range parameters once for each combination of choice
# parameters, then takes the optimum of those optima. The cost associated with this
# method grows with the number of combinations, and so it is only used when the
# number of enumerated discrete combinations is below some maximum value.
MAX_DISCRETE_ENUMERATIONS_MIXED = 65
MAX_DISCRETE_ENUMERATIONS_NO_CONTINUOUS_OPTIMIZATION = 1e4
MAX_ONE_HOT_ENCODINGS_CONTINUOUS_OPTIMIZATION = 33
SAASBO_INCOMPATIBLE_MESSAGE = (
    "SAASBO is incompatible with {} generation strategy. "
    "Disregarding user input `use_saasbo = True`."
)


def _make_sobol_step(
    num_trials: int = -1,
    min_trials_observed: Optional[int] = None,
    enforce_num_trials: bool = True,
    max_parallelism: Optional[int] = None,
    seed: Optional[int] = None,
    should_deduplicate: bool = False,
) -> GenerationStep:
    """Shortcut for creating a Sobol generation step."""
    return GenerationStep(
        model=Models.SOBOL,
        num_trials=num_trials,
        # NOTE: ceil(-1 / 2) = 0, so this is safe to do when num trials is -1.
        min_trials_observed=min_trials_observed or ceil(num_trials / 2),
        enforce_num_trials=enforce_num_trials,
        max_parallelism=max_parallelism,
        model_kwargs={"deduplicate": True, "seed": seed},
        should_deduplicate=should_deduplicate,
    )


def _make_botorch_step(
    num_trials: int = -1,
    min_trials_observed: Optional[int] = None,
    enforce_num_trials: bool = True,
    max_parallelism: Optional[int] = None,
    model: ModelRegistryBase = Models.BOTORCH_MODULAR,
    model_kwargs: Optional[Dict[str, Any]] = None,
    winsorization_config: Optional[
        Union[WinsorizationConfig, Dict[str, WinsorizationConfig]]
    ] = None,
    no_winsorization: bool = False,
    should_deduplicate: bool = False,
    verbose: Optional[bool] = None,
    disable_progbar: Optional[bool] = None,
    jit_compile: Optional[bool] = None,
    derelativize_with_raw_status_quo: bool = False,
    fit_out_of_design: bool = False,
) -> GenerationStep:
    """Shortcut for creating a BayesOpt generation step."""
    model_kwargs = model_kwargs or {}

    winsorization_transform_config = _get_winsorization_transform_config(
        winsorization_config=winsorization_config,
        no_winsorization=no_winsorization,
        derelativize_with_raw_status_quo=derelativize_with_raw_status_quo,
    )

    derelativization_transform_config = {
        "use_raw_status_quo": derelativize_with_raw_status_quo
    }
    model_kwargs["transform_configs"] = model_kwargs.get("transform_configs", {})
    model_kwargs["transform_configs"][
        "Derelativize"
    ] = derelativization_transform_config
    model_kwargs["fit_out_of_design"] = fit_out_of_design

    if not no_winsorization:
        _, default_bridge_kwargs = model.view_defaults()
        default_transforms = default_bridge_kwargs["transforms"]
        transforms = model_kwargs.get("transforms", default_transforms)
        model_kwargs["transforms"] = [cast(Type[Transform], Winsorize)] + transforms
        if winsorization_transform_config is not None:
            model_kwargs["transform_configs"][
                "Winsorize"
            ] = winsorization_transform_config

    if MODEL_KEY_TO_MODEL_SETUP[model.value].model_class != ModularBoTorchModel:
        if verbose is not None:
            model_kwargs.update({"verbose": verbose})
        if disable_progbar is not None:
            model_kwargs.update({"disable_progbar": disable_progbar})
        if jit_compile is not None:
            model_kwargs.update({"jit_compile": jit_compile})
    else:
        # TODO[T164389105] Rewrite choose_generation_strategy to be MBM first
        logger.info(
            "`verbose`, `disable_progbar`, and `jit_compile` are not yet supported "
            "when using `choose_generation_strategy` with ModularBoTorchModel, "
            "dropping these arguments."
        )
    return GenerationStep(
        model=model,
        num_trials=num_trials,
        # NOTE: ceil(-1 / 2) = 0, so this is safe to do when num trials is -1.
        min_trials_observed=min_trials_observed or ceil(num_trials / 2),
        enforce_num_trials=enforce_num_trials,
        max_parallelism=max_parallelism,
        model_kwargs=model_kwargs,
        should_deduplicate=should_deduplicate,
    )


def _suggest_gp_model(
    search_space: SearchSpace,
    num_trials: Optional[int] = None,
    optimization_config: Optional[OptimizationConfig] = None,
    use_saasbo: bool = False,
) -> Union[None, ModelRegistryBase]:
    """Suggest a model based on the search space. None means we use Sobol.

    1. We use Sobol if the number of total iterations in the optimization is
    known in advance and there are fewer distinct points in the search space
    than the known intended number of total iterations.
    2. We use ``BO_MIXED`` if there are fewer ordered parameters in the search space
    than the sum of options for the *unordered* choice parameters, and the number
    of discrete enumerations to be performed by the optimizer is less than
    ``MAX_DISCRETE_ENUMERATIONS_MIXED``, or if there are only choice parameters and
    the number of choice combinations to enumerate is less than
    ``MAX_DISCRETE_ENUMERATIONS_CHOICE_ONLY``. Note that we do not count 2-level choice
    parameters as unordered, since these do not affect the modeling choice.
    3. If there are more ordered parameters in the search space than the sum of options
    for the *unordered* choice parameters, or if there is at least one ordered
    parameter and the number of parameters needed to encode all unordered parameters
    is less than ``MAX_ONE_HOT_ENCODINGS_CONTINUOUS_OPTIMIZATION``, we use BO with
    continuous relaxation.
    * For BO, we use Modular BoTorch Model with ``SingleTaskGP`` if ``use_saasbo is
    False`` and with ``SaasFullyBayesianSingleTaskGP`` (aka SAASBO) otherwise.
    """
    # Count tunable parameter types.
    num_ordered_parameters = 0
    num_unordered_choices = 0
    num_enumerated_combinations = 1
    num_possible_points = 1
    all_range_parameters_are_discrete = True
    all_parameters_are_enumerated = True
    for parameter in search_space.tunable_parameters.values():
        should_enumerate_param = False
        num_param_discrete_values = None
        if isinstance(parameter, ChoiceParameter):
            should_enumerate_param = True
            num_param_discrete_values = len(parameter.values)
            num_possible_points *= num_param_discrete_values
            if parameter.is_ordered is False and num_param_discrete_values > 2:
                num_unordered_choices += num_param_discrete_values
            else:
                num_ordered_parameters += 1
        elif isinstance(parameter, RangeParameter):
            num_ordered_parameters += 1
            if parameter.parameter_type == ParameterType.FLOAT:
                all_range_parameters_are_discrete = False
            else:
                num_param_discrete_values = parameter.cardinality()
                num_possible_points *= num_param_discrete_values

        if should_enumerate_param:
            num_enumerated_combinations *= not_none(num_param_discrete_values)
        else:
            all_parameters_are_enumerated = False

    # Use Sobol if number of trials is known and sufficient to try all possible points.
    if (
        num_trials is not None
        and all_range_parameters_are_discrete
        and num_possible_points <= num_trials
    ):
        logger.info("Using Sobol since we can enumerate the search space.")
        if use_saasbo:
            logger.warning(SAASBO_INCOMPATIBLE_MESSAGE.format("Sobol"))
        return None

    # Use mixed Bayesian optimization when appropriate. This logic is currently tied to
    # the fact that acquisition function optimization for mixed BayesOpt currently
    # enumerates all combinations of choice parameters.
    # We use continuous relaxation if there are more ordered parameters than there
    # are choices for unordered parameters.
    if (num_ordered_parameters < num_unordered_choices) and (
        num_enumerated_combinations <= MAX_DISCRETE_ENUMERATIONS_MIXED
        or (
            all_parameters_are_enumerated
            and num_enumerated_combinations
            < MAX_DISCRETE_ENUMERATIONS_NO_CONTINUOUS_OPTIMIZATION
        )
    ):
        logger.info(
            "Using Bayesian optimization with a categorical kernel for improved "
            "performance with a large number of unordered categorical parameters."
        )
        if use_saasbo:
            logger.warning(SAASBO_INCOMPATIBLE_MESSAGE.format("`BO_MIXED`"))
        return Models.BO_MIXED

    if num_ordered_parameters >= num_unordered_choices or (
        num_unordered_choices < MAX_ONE_HOT_ENCODINGS_CONTINUOUS_OPTIMIZATION
        and num_ordered_parameters > 0
    ):
        # These use one-hot encoding for unordered choice parameters, resulting in a
        # total of num_unordered_choices OHE parameters.
        # So, we do not want to use them when there are too many unordered choices.
        method = Models.SAASBO if use_saasbo else Models.BOTORCH_MODULAR
        reason = (
            (
                "there are more ordered parameters than there are categories for the "
                "unordered categorical parameters."
                if num_ordered_parameters >= num_unordered_choices
                else "there is at least one ordered parameter and there are fewer than "
                f"{MAX_ONE_HOT_ENCODINGS_CONTINUOUS_OPTIMIZATION} choices for "
                "unordered parameters."
            )
            if num_unordered_choices > 0
            else "there is at least one ordered parameter"
            " and there are no unordered categorical parameters."
        )
        logger.info(f"Using {method} since {reason}")
        return method

    logger.info(
        f"Using Sobol since there are more than {MAX_DISCRETE_ENUMERATIONS_MIXED} "
        "combinations of enumerated parameters. For improved performance, make sure "
        "that all ordered `ChoiceParameter`s are encoded as such (`is_ordered=True`), "
        "and use `RangeParameter`s in place of ordered `ChoiceParameter`s where "
        "possible. Also, consider removing some or all unordered `ChoiceParameter`s."
    )
    if use_saasbo:
        logger.warning(SAASBO_INCOMPATIBLE_MESSAGE.format("Sobol"))
    return None


def calculate_num_initialization_trials(
    num_tunable_parameters: int,
    num_trials: Optional[int],
    use_batch_trials: bool,
) -> int:
    """
    Applies rules from high to low priority
     - 1 for batch trials.
     - At least 5
     - At most 1/5th of num_trials.
     - Twice the number of tunable parameters
    """
    if use_batch_trials:  # Batched trials.
        return 1

    ret = 2 * num_tunable_parameters
    if num_trials is not None:
        ret = min(ret, not_none(num_trials) // 5)
    return max(ret, 5)


def choose_generation_strategy(
    search_space: SearchSpace,
    *,
    use_batch_trials: bool = False,
    enforce_sequential_optimization: bool = True,
    random_seed: Optional[int] = None,
    torch_device: Optional[torch.device] = None,
    no_winsorization: bool = False,
    winsorization_config: Optional[
        Union[WinsorizationConfig, Dict[str, WinsorizationConfig]]
    ] = None,
    derelativize_with_raw_status_quo: bool = False,
    no_bayesian_optimization: bool = False,
    num_trials: Optional[int] = None,
    num_initialization_trials: Optional[int] = None,
    num_completed_initialization_trials: int = 0,
    max_initialization_trials: Optional[int] = None,
    min_sobol_trials_observed: Optional[int] = None,
    max_parallelism_cap: Optional[int] = None,
    max_parallelism_override: Optional[int] = None,
    optimization_config: Optional[OptimizationConfig] = None,
    should_deduplicate: bool = False,
    use_saasbo: bool = False,
    verbose: Optional[bool] = None,
    disable_progbar: Optional[bool] = None,
    jit_compile: Optional[bool] = None,
    experiment: Optional[Experiment] = None,
    suggested_model_override: Optional[ModelRegistryBase] = None,
    fit_out_of_design: bool = False,
) -> GenerationStrategy:
    """Select an appropriate generation strategy based on the properties of
    the search space and expected settings of the experiment, such as number of
    arms per trial, optimization algorithm settings, expected number of trials
    in the experiment, etc.

    Args:
        search_space: SearchSpace, based on the properties of which to select the
            generation strategy.
        use_batch_trials: Whether this generation strategy will be used to generate
            batched trials instead of 1-arm trials.
        enforce_sequential_optimization: Whether to enforce that 1) the generation
            strategy needs to be updated with ``min_trials_observed`` observations for
            a given generation step before proceeding to the next one and 2) maximum
            number of trials running at once (max_parallelism) if enforced for the
            BayesOpt step. NOTE: ``max_parallelism_override`` and
            ``max_parallelism_cap`` settings will still take their effect on max
            parallelism even if ``enforce_sequential_optimization=False``, so if those
            settings are specified, max parallelism will be enforced.
        random_seed: Fixed random seed for the Sobol generator.
        torch_device: The device to use for generation steps implemented in PyTorch
            (e.g. via BoTorch). Some generation steps (in particular EHVI-based ones
            for multi-objective optimization) can be sped up by running candidate
            generation on the GPU. If not specified, uses the default torch device
            (usually the CPU).
        no_winsorization: Whether to apply the winsorization transform
            prior to applying other transforms for fitting the BoTorch model.
        winsorization_config: Explicit winsorization settings, if winsorizing. Usually
            only `upper_quantile_margin` is set when minimizing, and only
            `lower_quantile_margin` when maximizing.
        derelativize_with_raw_status_quo: Whether to derelativize using the raw status
            quo values in any transforms. This argument is primarily to allow automatic
            Winsorization when relative constraints are present. Note: automatic
            Winsorization will fail if this is set to `False` (or unset) and there
            are relative constraints present.
        no_bayesian_optimization: If True, Bayesian optimization generation
            strategy will not be suggested and quasi-random strategy will be used.
        num_trials: Total number of trials in the optimization, if
            known in advance.
        num_initialization_trials: Specific number of initialization trials, if wanted.
            Typically, initialization trials are generated quasi-randomly.
        max_initialization_trials: If ``num_initialization_trials`` unspecified, it
            will be determined automatically. This arg provides a cap on that
            automatically determined number.
        num_completed_initialization_trials: The final calculated number of
            initialization trials is reduced by this number. This is useful when
            warm-starting an experiment, to specify what number of completed trials
            can be used to satisfy the initialization_trial requirement.
        min_sobol_trials_observed: Minimum number of Sobol trials that must be
            observed before proceeding to the next generation step. Defaults to
            `ceil(num_initialization_trials / 2)`.
        max_parallelism_cap: Integer cap on parallelism in this generation strategy.
            If specified, ``max_parallelism`` setting in each generation step will be
            set to the minimum of the default setting for that step and the value of
            this cap. ``max_parallelism_cap`` is meant to just be a hard limit on
            parallelism (e.g. to avoid overloading machine(s) that evaluate the
            experiment trials). Specify only if not specifying
            ``max_parallelism_override``.
        max_parallelism_override: Integer, with which to override the default max
            parallelism setting for all steps in the generation strategy returned from
            this function. Each generation step has a ``max_parallelism`` value, which
            restricts how many trials can run simultaneously during a given generation
            step. By default, the parallelism setting is chosen as appropriate for the
            model in a given generation step. If ``max_parallelism_override`` is -1,
            no max parallelism will be enforced for any step of the generation
            strategy. Be aware that parallelism is limited to improve performance of
            Bayesian optimization, so only disable its limiting if necessary.
        optimization_config: used to infer whether to use MOO and will be passed in to
            ``Winsorize`` via its ``transform_config`` in order to determine default
            winsorization behavior when necessary.
        should_deduplicate: Whether to deduplicate the parameters of proposed arms
            against those of previous arms via rejection sampling. If this is True,
            the generation strategy will discard generator runs produced from the
            generation step that has `should_deduplicate=True` if they contain arms
            already present on the experiment and replace them with new generator runs.
            If no generator run with entirely unique arms could be produced in 5
            attempts, a `GenerationStrategyRepeatedPoints` error will be raised, as we
            assume that the optimization converged when the model can no longer suggest
            unique arms.
        use_saasbo: Whether to use SAAS prior for any GPEI generation steps.
        verbose: Whether GP model should produce verbose logs. If not ``None``, its
            value gets added to ``model_kwargs`` during ``generation_strategy``
            construction. Defaults to ``True`` for SAASBO, else ``None``. Verbose
            outputs are currently only available for SAASBO, so if ``verbose is not
            None`` for a different model type, it will be overridden to ``None`` with
            a warning.
        disable_progbar: Whether GP model should produce a progress bar. If not
            ``None``, its value gets added to ``model_kwargs`` during
            ``generation_strategy`` construction. Defaults to ``True`` for SAASBO, else
            ``None``. Progress bars are currently only available for SAASBO, so if
            ``disable_probar is not None`` for a different model type, it will be
            overridden to ``None`` with a warning.
        jit_compile: Whether to use jit compilation in Pyro when SAASBO is used.
        experiment: If specified, ``_experiment`` attribute of the generation strategy
            will be set to this experiment (useful for associating a generation
            strategy with a given experiment before it's first used to ``gen`` with
            that experiment). Can also provide `optimization_config` if it is not
            provided as an arg to this function.
        suggested_model_override: If specified, this model will be used for the GP
            step and automatic selection will be skipped.
        fit_out_of_design: Whether to include out-of-design points in the model.
    """
    if experiment is not None and optimization_config is None:
        optimization_config = experiment.optimization_config

    suggested_model = suggested_model_override or _suggest_gp_model(
        search_space=search_space,
        num_trials=num_trials,
        optimization_config=optimization_config,
        use_saasbo=use_saasbo,
    )
    # Determine max parallelism for the generation steps.
    if max_parallelism_override == -1:
        # `max_parallelism_override` of -1 means no max parallelism enforcement in
        # the generation strategy, which means `max_parallelism=None` in gen. steps.
        sobol_parallelism = bo_parallelism = None
    elif max_parallelism_override is not None:
        sobol_parallelism = bo_parallelism = max_parallelism_override
    elif max_parallelism_cap is not None:  # Max parallelism override is None by now
        sobol_parallelism = max_parallelism_cap
        bo_parallelism = min(max_parallelism_cap, DEFAULT_BAYESIAN_PARALLELISM)
    elif not enforce_sequential_optimization:
        # If no max parallelism settings specified and not enforcing sequential
        # optimization, do not limit parallelism.
        sobol_parallelism = bo_parallelism = None
    else:  # No additional max parallelism settings, use defaults
        sobol_parallelism = None  # No restriction on Sobol phase
        bo_parallelism = DEFAULT_BAYESIAN_PARALLELISM

    if not no_bayesian_optimization and suggested_model is not None:
        if not enforce_sequential_optimization and (
            max_parallelism_override or max_parallelism_cap
        ):
            logger.info(
                "If `enforce_sequential_optimization` is False, max parallelism is "
                "not enforced and other max parallelism settings will be ignored."
            )
        if max_parallelism_override and max_parallelism_cap:
            raise ValueError(
                "If `max_parallelism_override` specified, cannot also apply "
                "`max_parallelism_cap`."
            )

        # If number of initialization trials is not specified, estimate it.
        logger.info(
            "Calculating the number of remaining initialization trials based on "
            f"num_initialization_trials={num_initialization_trials} "
            f"max_initialization_trials={max_initialization_trials} "
            f"num_tunable_parameters={len(search_space.tunable_parameters)} "
            f"num_trials={num_trials} "
            f"use_batch_trials={use_batch_trials}"
        )
        if num_initialization_trials is None:
            num_initialization_trials = calculate_num_initialization_trials(
                num_tunable_parameters=len(search_space.tunable_parameters),
                num_trials=num_trials,
                use_batch_trials=use_batch_trials,
            )
            if max_initialization_trials is not None:
                num_initialization_trials = min(
                    num_initialization_trials, max_initialization_trials
                )
            logger.info(
                f"calculated num_initialization_trials={num_initialization_trials}"
            )
        num_remaining_initialization_trials = max(
            0, num_initialization_trials - max(0, num_completed_initialization_trials)
        )
        logger.info(
            "num_completed_initialization_trials="
            f"{num_completed_initialization_trials} "
            f"num_remaining_initialization_trials={num_remaining_initialization_trials}"
        )
        steps = []
        # `verbose` and `disable_progbar` defaults and overrides
        model_is_saasbo = suggested_model is Models.SAASBO
        if verbose is None and model_is_saasbo:
            verbose = True
        elif verbose is not None and not model_is_saasbo:
            logger.warning(
                f"Overriding `verbose = {verbose}` to `None` for non-SAASBO GP step."
            )
            verbose = None
        if disable_progbar is not None and not model_is_saasbo:
            logger.warning(
                f"Overriding `disable_progbar = {disable_progbar}` to `None` for "
                "non-SAASBO GP step."
            )
            disable_progbar = None
        if jit_compile is not None and not model_is_saasbo:
            logger.warning(
                f"Overriding `jit_compile = {jit_compile}` to `None` for "
                "non-SAASBO GP step."
            )
            jit_compile = None

        model_kwargs: Dict[str, Any] = {
            "torch_device": torch_device,
            "fit_out_of_design": fit_out_of_design,
        }

        # Create `generation_strategy`, adding first Sobol step
        # if `num_remaining_initialization_trials` is > 0.
        if num_remaining_initialization_trials > 0:
            steps.append(
                _make_sobol_step(
                    num_trials=num_remaining_initialization_trials,
                    min_trials_observed=min_sobol_trials_observed,
                    enforce_num_trials=enforce_sequential_optimization,
                    seed=random_seed,
                    max_parallelism=sobol_parallelism,
                    should_deduplicate=should_deduplicate,
                )
            )
        steps.append(
            _make_botorch_step(
                model=suggested_model,
                winsorization_config=winsorization_config,
                derelativize_with_raw_status_quo=derelativize_with_raw_status_quo,
                no_winsorization=no_winsorization,
                max_parallelism=bo_parallelism,
                model_kwargs=model_kwargs,
                should_deduplicate=should_deduplicate,
                verbose=verbose,
                disable_progbar=disable_progbar,
                jit_compile=jit_compile,
            ),
        )
        gs = GenerationStrategy(steps=steps)
        logger.info(
            f"Using Bayesian Optimization generation strategy: {gs}. Iterations after"
            f" {num_remaining_initialization_trials} will take longer to generate due"
            " to model-fitting."
        )
    else:  # `no_bayesian_optimization` is True or we could not suggest BO model
        if verbose is not None:
            logger.warning(
                f"Ignoring `verbose = {verbose}` for `generation_strategy` "
                "without a GP step."
            )

        gs = GenerationStrategy(
            steps=[
                _make_sobol_step(
                    seed=random_seed,
                    should_deduplicate=should_deduplicate,
                    max_parallelism=sobol_parallelism,
                )
            ]
        )
        logger.info("Using Sobol generation strategy.")
    if experiment:
        gs.experiment = experiment
    return gs


def _get_winsorization_transform_config(
    winsorization_config: Optional[
        Union[WinsorizationConfig, Dict[str, WinsorizationConfig]]
    ],
    derelativize_with_raw_status_quo: bool,
    no_winsorization: bool,
) -> Optional[TConfig]:
    if no_winsorization:
        if winsorization_config is not None:
            warnings.warn(
                "`no_winsorization = True` but `winsorization_config` has been set. "
                "Not winsorizing.",
                stacklevel=2,
            )
        return None
    if winsorization_config:
        return {"winsorization_config": winsorization_config}
    return {"derelativize_with_raw_status_quo": derelativize_with_raw_status_quo}



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import dataclasses
from typing import Any, Dict
from unittest import mock

import torch
from ax.core.search_space import SearchSpaceDigest
from ax.models.torch.botorch_kg import _instantiate_KG, KnowledgeGradient
from ax.models.torch_base import TorchOptConfig
from ax.utils.common.testutils import TestCase
from ax.utils.testing.mock import fast_botorch_optimize
from botorch.acquisition.analytic import PosteriorMean
from botorch.acquisition.fixed_feature import FixedFeatureAcquisitionFunction
from botorch.acquisition.knowledge_gradient import qMultiFidelityKnowledgeGradient
from botorch.acquisition.monte_carlo import qSimpleRegret
from botorch.acquisition.objective import (
    LinearMCObjective,
    ScalarizedPosteriorTransform,
)
from botorch.exceptions.errors import UnsupportedError
from botorch.models.transforms.input import Warp
from botorch.sampling.normal import IIDNormalSampler, SobolQMCNormalSampler
from botorch.utils.datasets import SupervisedDataset


def dummy_func(X: torch.Tensor) -> torch.Tensor:
    return X


class KnowledgeGradientTest(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.tkwargs: Dict[str, Any] = {
            "device": torch.device("cpu"),
            "dtype": torch.double,
        }
        self.feature_names = ["x1", "x2", "x3"]
        self.metric_names = ["y"]
        self.dataset = SupervisedDataset(
            X=torch.tensor([[1.0, 2.0, 3.0], [2.0, 3.0, 4.0]], **self.tkwargs),
            Y=torch.tensor([[3.0], [4.0]], **self.tkwargs),
            Yvar=torch.tensor([[0.0], [2.0]], **self.tkwargs),
            feature_names=self.feature_names,
            outcome_names=self.metric_names,
        )
        self.bounds = [(0.0, 1.0), (1.0, 4.0), (2.0, 5.0)]
        self.acq_options = {"num_fantasies": 30, "mc_samples": 30}
        self.objective_weights = torch.tensor([1.0], **self.tkwargs)
        self.optimizer_options = {
            "num_restarts": 12,
            "raw_samples": 12,
            "options": {
                "maxiter": 5,
                "batch_limit": 1,
            },
        }
        self.optimize_acqf = "ax.models.torch.botorch_kg.optimize_acqf"
        self.X_dummy = torch.ones(1, 3, **self.tkwargs)
        self.outcome_constraints = (torch.tensor([[1.0]]), torch.tensor([[0.5]]))
        self.objective_weights = torch.ones(1, **self.tkwargs)
        self.moo_objective_weights = torch.ones(2, **self.tkwargs)
        self.objective_thresholds = torch.tensor([0.5, 1.5])
        self.search_space_digest = SearchSpaceDigest(
            feature_names=self.feature_names,
            bounds=self.bounds,
        )

    @fast_botorch_optimize
    def test_KnowledgeGradient(self) -> None:
        model = KnowledgeGradient()
        model.fit(
            datasets=[self.dataset],
            search_space_digest=self.search_space_digest,
        )

        n = 2

        X_dummy = torch.rand(1, n, 4, **self.tkwargs)
        acq_dummy = torch.tensor(0.0, **self.tkwargs)

        torch_opt_config = TorchOptConfig(
            objective_weights=self.objective_weights,
            model_gen_options={
                "acquisition_function_kwargs": self.acq_options,
                "optimizer_kwargs": self.optimizer_options,
            },
        )

        with mock.patch(self.optimize_acqf) as mock_optimize_acqf:
            mock_optimize_acqf.side_effect = [(X_dummy, acq_dummy)]
            gen_results = model.gen(
                n=n,
                search_space_digest=self.search_space_digest,
                torch_opt_config=torch_opt_config,
            )
            self.assertTrue(torch.equal(gen_results.points, X_dummy.cpu()))
            self.assertTrue(
                torch.equal(
                    gen_results.weights, torch.ones(n, dtype=self.tkwargs["dtype"])
                )
            )

            # called once, the best point call is not caught by mock
            mock_optimize_acqf.assert_called_once()

        ini_dummy = torch.rand(10, 32, 3, **self.tkwargs)
        optimizer_options2 = {
            "num_restarts": 1,
            "raw_samples": 1,
            "options": {
                "maxiter": 5,
                "batch_limit": 1,
            },
            "partial_restarts": 2,
        }
        torch_opt_config.model_gen_options["optimizer_kwargs"] = optimizer_options2
        with mock.patch(
            "ax.models.torch.botorch_kg.gen_one_shot_kg_initial_conditions",
            return_value=ini_dummy,
        ) as mock_warmstart_initialization:
            gen_results = model.gen(
                n=n,
                search_space_digest=self.search_space_digest,
                torch_opt_config=torch_opt_config,
            )
            mock_warmstart_initialization.assert_called_once()

        posterior_tf = ScalarizedPosteriorTransform(weights=self.objective_weights)
        dummy_acq = PosteriorMean(model=model.model, posterior_transform=posterior_tf)
        with mock.patch(
            "ax.models.torch.utils.PosteriorMean", return_value=dummy_acq
        ) as mock_posterior_mean, mock.patch(
            "ax.models.torch.utils.get_botorch_objective_and_transform",
            return_value=(
                None,
                ScalarizedPosteriorTransform(
                    weights=torch.tensor([1.0], dtype=torch.double)
                ),
            ),
        ):
            gen_results = model.gen(
                n=n,
                search_space_digest=self.search_space_digest,
                torch_opt_config=torch_opt_config,
            )
            self.assertEqual(mock_posterior_mean.call_count, 2)

        # Check best point selection within bounds (some numerical tolerance)
        xbest = model.best_point(
            search_space_digest=self.search_space_digest,
            torch_opt_config=torch_opt_config,
        )
        lb = torch.tensor([b[0] for b in self.bounds]) - 1e-5
        ub = torch.tensor([b[1] for b in self.bounds]) + 1e-5
        self.assertTrue(torch.all(xbest <= ub))
        self.assertTrue(torch.all(xbest >= lb))

        # test error message
        torch_opt_config = dataclasses.replace(
            torch_opt_config,
            linear_constraints=(
                torch.tensor([[0.0, 0.0, 0.0], [0.0, 1.0, 0.0]]),
                torch.tensor([[0.5], [1.0]]),
            ),
        )
        with self.assertRaises(UnsupportedError):
            gen_results = model.gen(
                n=n,
                search_space_digest=self.search_space_digest,
                torch_opt_config=torch_opt_config,
            )

        # test input warping
        self.assertFalse(model.use_input_warping)
        model = KnowledgeGradient(use_input_warping=True)
        model.fit(
            datasets=[self.dataset],
            search_space_digest=self.search_space_digest,
        )
        self.assertTrue(model.use_input_warping)
        self.assertTrue(hasattr(model.model, "input_transform"))
        self.assertIsInstance(model.model.input_transform, Warp)

        # test loocv pseudo likelihood
        self.assertFalse(model.use_loocv_pseudo_likelihood)
        model = KnowledgeGradient(use_loocv_pseudo_likelihood=True)
        model.fit(
            datasets=[self.dataset],
            search_space_digest=self.search_space_digest,
        )
        self.assertTrue(model.use_loocv_pseudo_likelihood)

    @fast_botorch_optimize
    def test_KnowledgeGradient_multifidelity(self) -> None:
        search_space_digest = SearchSpaceDigest(
            feature_names=self.feature_names,
            bounds=self.bounds,
            fidelity_features=[2],
            target_values={2: 5.0},
        )
        model = KnowledgeGradient()
        model.fit(
            datasets=[self.dataset],
            search_space_digest=search_space_digest,
        )

        torch_opt_config = TorchOptConfig(
            objective_weights=self.objective_weights,
            model_gen_options={
                "acquisition_function_kwargs": self.acq_options,
                "optimizer_kwargs": self.optimizer_options,
            },
        )
        # Check best point selection within bounds (some numerical tolerance)
        xbest = model.best_point(
            search_space_digest=search_space_digest,
            torch_opt_config=torch_opt_config,
        )
        lb = torch.tensor([b[0] for b in self.bounds]) - 1e-5
        ub = torch.tensor([b[1] for b in self.bounds]) + 1e-5
        self.assertTrue(torch.all(xbest <= ub))
        self.assertTrue(torch.all(xbest >= lb))

        # check error when no target fidelities are specified
        with self.assertRaises(RuntimeError):
            model.best_point(
                search_space_digest=dataclasses.replace(
                    search_space_digest,
                    target_values={},
                ),
                torch_opt_config=torch_opt_config,
            )

        # check generation
        n = 2
        X_dummy = torch.zeros(1, n, 3, **self.tkwargs)
        acq_dummy = torch.tensor(0.0, **self.tkwargs)
        dummy = (X_dummy, acq_dummy)
        with mock.patch(self.optimize_acqf, side_effect=[dummy]) as mock_optimize_acqf:
            gen_results = model.gen(
                n=n,
                search_space_digest=search_space_digest,
                torch_opt_config=torch_opt_config,
            )
            self.assertTrue(torch.equal(gen_results.points, X_dummy.cpu()))
            self.assertTrue(
                torch.equal(
                    gen_results.weights, torch.ones(n, dtype=self.tkwargs["dtype"])
                )
            )
            mock_optimize_acqf.assert_called()  # called twice, once for best_point

        # test error message
        with self.assertRaises(UnsupportedError):
            xbest = model.best_point(
                search_space_digest=search_space_digest,
                torch_opt_config=dataclasses.replace(
                    torch_opt_config,
                    linear_constraints=(
                        torch.tensor([[0.0, 0.0, 0.0], [0.0, 1.0, 0.0]]),
                        torch.tensor([[0.5], [1.0]]),
                    ),
                ),
            )

        # test input warping
        self.assertFalse(model.use_input_warping)
        model = KnowledgeGradient(use_input_warping=True)
        model.fit(
            datasets=[self.dataset],
            search_space_digest=search_space_digest,
        )
        self.assertTrue(model.use_input_warping)
        self.assertTrue(hasattr(model.model, "input_transform"))
        self.assertIsInstance(model.model.input_transform, Warp)

        # test loocv pseudo likelihood
        self.assertFalse(model.use_loocv_pseudo_likelihood)
        model = KnowledgeGradient(use_loocv_pseudo_likelihood=True)
        model.fit(
            datasets=[self.dataset],
            search_space_digest=search_space_digest,
        )
        self.assertTrue(model.use_loocv_pseudo_likelihood)

    @fast_botorch_optimize
    def test_KnowledgeGradient_helpers(self) -> None:
        model = KnowledgeGradient()
        model.fit(
            datasets=[self.dataset],
            search_space_digest=SearchSpaceDigest(
                feature_names=self.feature_names,
                bounds=self.bounds,
            ),
        )

        # test _instantiate_KG
        posterior_tf = ScalarizedPosteriorTransform(weights=self.objective_weights)

        # test acquisition setting
        acq_function = _instantiate_KG(
            model=model.model,
            posterior_transform=posterior_tf,
            n_fantasies=10,
            qmc=True,
        )
        self.assertIsInstance(acq_function.sampler, SobolQMCNormalSampler)
        self.assertIsInstance(
            acq_function.posterior_transform, ScalarizedPosteriorTransform
        )
        self.assertEqual(acq_function.num_fantasies, 10)

        acq_function = _instantiate_KG(
            model=model.model,
            posterior_transform=posterior_tf,
            n_fantasies=10,
            qmc=False,
        )
        self.assertIsInstance(acq_function.sampler, IIDNormalSampler)

        acq_function = _instantiate_KG(
            model=model.model,
            posterior_transform=posterior_tf,
            qmc=False,
        )
        self.assertIsNone(acq_function.inner_sampler)

        acq_function = _instantiate_KG(
            model=model.model,
            posterior_transform=posterior_tf,
            qmc=True,
            X_pending=self.X_dummy,
        )
        self.assertIsNone(acq_function.inner_sampler)
        self.assertTrue(torch.equal(acq_function.X_pending, self.X_dummy))

        # test _get_best_point_acqf
        acq_function, non_fixed_idcs = model._get_best_point_acqf(
            objective_weights=self.objective_weights,
            outcome_constraints=self.outcome_constraints,
            X_observed=self.X_dummy,
        )
        self.assertIsInstance(acq_function, qSimpleRegret)
        self.assertIsInstance(acq_function.sampler, SobolQMCNormalSampler)
        self.assertIsNone(non_fixed_idcs)

        acq_function, non_fixed_idcs = model._get_best_point_acqf(
            objective_weights=self.objective_weights,
            outcome_constraints=self.outcome_constraints,
            X_observed=self.X_dummy,
            qmc=False,
        )
        self.assertIsInstance(acq_function.sampler, IIDNormalSampler)
        self.assertIsNone(non_fixed_idcs)

        with self.assertRaises(RuntimeError):
            model._get_best_point_acqf(
                objective_weights=self.objective_weights,
                outcome_constraints=self.outcome_constraints,
                X_observed=self.X_dummy,
                target_fidelities={1: 1.0},
            )

        # multi-fidelity tests

        model = KnowledgeGradient()
        model.fit(
            datasets=[self.dataset],
            search_space_digest=SearchSpaceDigest(
                feature_names=self.feature_names,
                bounds=self.bounds,
                fidelity_features=[-1],
            ),
        )

        acq_function = _instantiate_KG(
            model=model.model,
            posterior_transform=posterior_tf,
            target_fidelities={2: 1.0},
            # pyre-fixme[6]: For 4th param expected `Optional[Tensor]` but got `int`.
            current_value=0,
        )
        self.assertIsInstance(acq_function, qMultiFidelityKnowledgeGradient)

        acq_function = _instantiate_KG(
            model=model.model,
            objective=LinearMCObjective(weights=self.objective_weights),
        )
        self.assertIsInstance(acq_function.inner_sampler, SobolQMCNormalSampler)

        # test error that target fidelity and fidelity weight indices must match
        with self.assertRaises(RuntimeError):
            _instantiate_KG(
                model=model.model,
                posterior_transform=posterior_tf,
                target_fidelities={1: 1.0},
                fidelity_weights={2: 1.0},
                # pyre-fixme[6]: For 5th param expected `Optional[Tensor]` but got
                #  `int`.
                current_value=0,
            )

        # test _get_best_point_acqf
        acq_function, non_fixed_idcs = model._get_best_point_acqf(
            objective_weights=self.objective_weights,
            outcome_constraints=self.outcome_constraints,
            X_observed=self.X_dummy,
            target_fidelities={2: 1.0},
        )
        self.assertIsInstance(acq_function, FixedFeatureAcquisitionFunction)
        self.assertIsInstance(acq_function.acq_func.sampler, SobolQMCNormalSampler)
        self.assertEqual(non_fixed_idcs, [0, 1])

        acq_function, non_fixed_idcs = model._get_best_point_acqf(
            objective_weights=self.objective_weights,
            outcome_constraints=self.outcome_constraints,
            X_observed=self.X_dummy,
            target_fidelities={2: 1.0},
            qmc=False,
        )
        self.assertIsInstance(acq_function, FixedFeatureAcquisitionFunction)
        self.assertIsInstance(acq_function.acq_func.sampler, IIDNormalSampler)
        self.assertEqual(non_fixed_idcs, [0, 1])

        # test error that fixed features are provided
        with self.assertRaises(RuntimeError):
            model._get_best_point_acqf(
                objective_weights=self.objective_weights,
                outcome_constraints=self.outcome_constraints,
                X_observed=self.X_dummy,
                qmc=False,
            )

        # test error if fixed features are also fidelity features
        with self.assertRaises(RuntimeError):
            model._get_best_point_acqf(
                objective_weights=self.objective_weights,
                outcome_constraints=self.outcome_constraints,
                X_observed=self.X_dummy,
                fixed_features={2: 2.0},
                target_fidelities={2: 1.0},
                qmc=False,
            )

        # TODO: Test subsetting multi-output model
